# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from testtools import matchers

from paunch import runner
from paunch.tests import base


class TestBaseRunner(base.TestCase):
    def setUp(self):
        super(TestBaseRunner, self).setUp()
        self.runner = runner.DockerRunner('tester')
        self.podman_runner = runner.PodmanRunner('tester')

    def mock_execute(self, popen, stdout, stderr, returncode):
        subproc = mock.Mock()
        subproc.returncode = returncode
        subproc.communicate.return_value = (stdout.encode('utf-8'),
                                            stderr.encode('utf-8'))
        popen.return_value = subproc

    def assert_execute(self, popen, cmd):
        popen.assert_called_with(cmd, stderr=-1, stdout=-1)

    @mock.patch('subprocess.Popen')
    def test_execute(self, popen):
        self.mock_execute(popen, 'The stdout', 'The stderr', 0)

        self.assertEqual(
            ('The stdout', 'The stderr', 0),
            self.runner.execute(['ls', '-l'])
        )
        self.assert_execute(popen, ['ls', '-l'])

    @mock.patch('subprocess.Popen')
    def test_current_config_ids_docker(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)

        self.assertEqual(
            set(('one', 'two', 'three')),
            self.runner.current_config_ids()
        )
        self.assert_execute(
            popen, ['docker', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Label "config_id"}}']
        )

    @mock.patch('subprocess.Popen')
    def test_current_config_ids_podman(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)
        self.assertEqual(
            set(('one', 'two', 'three')),
            self.podman_runner.current_config_ids()
        )
        self.assert_execute(
            popen, ['podman', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Labels.config_id}}']
        )

    @mock.patch('subprocess.Popen')
    def test_containers_in_config(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)
        self.runner.remove_container = mock.Mock()

        result = self.runner.containers_in_config('foo')

        self.assert_execute(
            popen, ['docker', 'ps', '-q', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--filter', 'label=config_id=foo']
        )
        self.assertEqual(['one', 'two', 'three'], result)

    @mock.patch('subprocess.Popen')
    def test_inspect(self, popen):
        self.mock_execute(popen, '[{"foo": "bar"}]', '', 0)

        self.assertEqual(
            {"foo": "bar"},
            self.runner.inspect('one')
        )
        self.assert_execute(
            popen, ['docker', 'inspect', '--type', 'container', 'one']
        )

    @mock.patch('subprocess.Popen')
    def test_inspect_format(self, popen):
        self.mock_execute(popen, 'bar', '', 0)

        self.assertEqual(
            "bar",
            self.runner.inspect('one', output_format='{{foo}}')
        )
        self.assert_execute(
            popen, ['docker', 'inspect', '--type', 'container',
                    '--format', '{{foo}}', 'one']
        )

    def test_unique_container_name(self):
        self.runner.inspect = mock.Mock()
        self.runner.inspect.return_value = None
        self.assertEqual('foo', self.runner.unique_container_name('foo'))

        self.runner.inspect.side_effect = ['exists', 'exists', None]
        name = self.runner.unique_container_name('foo')
        name_pattern = '^foo-[a-z0-9]{8}$'
        self.assertThat(name, matchers.MatchesRegex(name_pattern))

    @mock.patch('subprocess.Popen')
    def test_discover_container_name(self, popen):
        self.mock_execute(popen, 'one-12345678', '', 0)

        self.assertEqual(
            'one-12345678',
            self.runner.discover_container_name('one', 'foo')
        )

        self.assert_execute(
            popen, ['docker', 'ps', '-a',
                    '--filter', 'label=container_name=one',
                    '--filter', 'label=config_id=foo',
                    '--format', '{{.Names}}']
        )

    @mock.patch('subprocess.Popen')
    def test_discover_container_name_empty(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.assertEqual(
            None,
            self.runner.discover_container_name('one', 'foo')
        )

    @mock.patch('subprocess.Popen')
    def test_discover_container_name_error(self, popen):
        self.mock_execute(popen, '', 'ouch', 1)

        self.assertEqual(
            None,
            self.runner.discover_container_name('one', 'foo')
        )

    @mock.patch('subprocess.Popen')
    def test_delete_missing_configs_docker(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree\nfour', '', 0)
        self.runner.remove_containers = mock.Mock()

        self.runner.delete_missing_configs(['two', 'three'])
        self.assert_execute(
            popen, ['docker', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Label "config_id"}}']
        )

        # containers one and four will be deleted
        self.runner.remove_containers.assert_has_calls([
            mock.call('one'), mock.call('four')
        ], any_order=True)

    @mock.patch('subprocess.Popen')
    def test_delete_missing_configs_podman(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree\nfour', '', 0)
        self.podman_runner.remove_containers = mock.Mock()

        self.podman_runner.delete_missing_configs(['two', 'three'])
        self.assert_execute(
            popen, ['podman', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Labels.config_id}}']
        )

        # containers one and four will be deleted
        self.podman_runner.remove_containers.assert_has_calls([
            mock.call('one'), mock.call('four')
        ], any_order=True)

    @mock.patch('subprocess.Popen')
    def test_list_configs_docker(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)
        self.runner.inspect = mock.Mock(
            return_value={'e': 'f'})
        self.runner.containers_in_config = mock.Mock(
            return_value=['a', 'b', 'c'])

        result = self.runner.list_configs()

        self.assert_execute(
            popen, ['docker', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Label "config_id"}}']
        )
        self.runner.containers_in_config.assert_has_calls([
            mock.call('one'), mock.call('two'), mock.call('three')
        ], any_order=True)
        self.runner.inspect.assert_has_calls([
            mock.call('a'), mock.call('b'), mock.call('c'),
            mock.call('a'), mock.call('b'), mock.call('c'),
            mock.call('a'), mock.call('b'), mock.call('c')
        ])
        self.assertEqual({
            'one': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}],
            'two': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}],
            'three': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}]
        }, result)

    @mock.patch('subprocess.Popen')
    def test_list_configs_podman(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)
        self.podman_runner.inspect = mock.Mock(
            return_value={'e': 'f'})
        self.podman_runner.containers_in_config = mock.Mock(
            return_value=['a', 'b', 'c'])

        result = self.podman_runner.list_configs()

        self.assert_execute(
            popen, ['podman', 'ps', '-a', '--filter',
                    'label=managed_by=tester',
                    '--format', '{{.Labels.config_id}}']
        )
        self.podman_runner.containers_in_config.assert_has_calls([
            mock.call('one'), mock.call('two'), mock.call('three')
        ], any_order=True)
        self.podman_runner.inspect.assert_has_calls([
            mock.call('a'), mock.call('b'), mock.call('c'),
            mock.call('a'), mock.call('b'), mock.call('c'),
            mock.call('a'), mock.call('b'), mock.call('c')
        ])
        self.assertEqual({
            'one': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}],
            'two': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}],
            'three': [{'e': 'f'}, {'e': 'f'}, {'e': 'f'}]
        }, result)

    @mock.patch('subprocess.Popen')
    def test_remove_containers(self, popen):
        self.mock_execute(popen, 'one\ntwo\nthree', '', 0)
        self.runner.remove_container = mock.Mock()

        self.runner.remove_containers('foo')

        self.assert_execute(
            popen, ['docker', 'ps', '-q', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--filter', 'label=config_id=foo']
        )
        self.runner.remove_container.assert_has_calls([
            mock.call('one'), mock.call('two'), mock.call('three')
        ])

    @mock.patch('subprocess.Popen')
    def test_remove_container(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.runner.remove_container('one')
        self.assert_execute(
            popen, ['docker', 'rm', 'one']
        )

    @mock.patch('subprocess.Popen')
    def test_stop_container(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.runner.stop_container('one')
        self.assert_execute(
            popen, ['docker', 'stop', 'one']
        )

    @mock.patch('subprocess.Popen')
    def test_stop_container_override(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.runner.stop_container('one', 'podman')
        self.assert_execute(
            popen, ['podman', 'stop', 'one']
        )

    @mock.patch('subprocess.Popen')
    def test_container_names_docker(self, popen):
        ps_result = '''one one
two-12345678 two
two two
three-12345678 three
four-12345678 four
'''

        self.mock_execute(popen, ps_result, '', 0)

        names = list(self.runner.container_names())

        self.assert_execute(
            popen, ['docker', 'ps', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--format', '{{.Names}} {{.Label "container_name"}}']
        )
        self.assertEqual([
            ['one', 'one'],
            ['two-12345678', 'two'],
            ['two', 'two'],
            ['three-12345678', 'three'],
            ['four-12345678', 'four']
        ], names)

    @mock.patch('subprocess.Popen')
    def test_container_names_podman(self, popen):
        ps_result = '''one one
two-12345678 two
two two
three-12345678 three
four-12345678 four
'''

        self.mock_execute(popen, ps_result, '', 0)

        names = list(self.podman_runner.container_names())

        self.assert_execute(
            popen, ['podman', 'ps', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--format', '{{.Names}} {{.Labels.container_name}}']
        )
        self.assertEqual([
            ['one', 'one'],
            ['two-12345678', 'two'],
            ['two', 'two'],
            ['three-12345678', 'three'],
            ['four-12345678', 'four']
        ], names)

    @mock.patch('subprocess.Popen')
    def test_container_names_by_conf_id_docker(self, popen):
        ps_result = '''one one
two-12345678 two
'''

        self.mock_execute(popen, ps_result, '', 0)

        names = list(self.runner.container_names('abc'))

        self.assert_execute(
            popen, ['docker', 'ps', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--filter', 'label=config_id=abc',
                    '--format', '{{.Names}} {{.Label "container_name"}}']
        )
        self.assertEqual([
            ['one', 'one'],
            ['two-12345678', 'two']
        ], names)

    @mock.patch('subprocess.Popen')
    def test_container_names_by_conf_id_podman(self, popen):
        ps_result = '''one one
two-12345678 two
'''

        self.mock_execute(popen, ps_result, '', 0)

        names = list(self.podman_runner.container_names('abc'))

        self.assert_execute(
            popen, ['podman', 'ps', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--filter', 'label=config_id=abc',
                    '--format', '{{.Names}} {{.Labels.container_name}}']
        )
        self.assertEqual([
            ['one', 'one'],
            ['two-12345678', 'two']
        ], names)

    @mock.patch('os.path.exists', return_value=True)
    def test_validate_volume_source_file(self, exists_mock):
        self.assertTrue(self.podman_runner.validate_volume_source('/tmp'))

    @mock.patch('os.path.exists', return_value=False)
    def test_validate_volume_source_file_fail(self, exists_mock):
        self.assertFalse(self.podman_runner.validate_volume_source('/nope'))

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('subprocess.Popen')
    def test_validate_volume_source_container(self, popen, exists_mock):
        ps_result = '''foo
foobar
'''
        self.mock_execute(popen, ps_result, '', 0)
        self.assertTrue(self.podman_runner.validate_volume_source('foo'))

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('subprocess.Popen')
    def test_validate_volume_source_container_fail(self, popen, exists_mock):
        ps_result = ''
        self.mock_execute(popen, ps_result, '', 0)
        self.assertFalse(self.podman_runner.validate_volume_source('foo'))


class TestDockerRunner(TestBaseRunner):

    @mock.patch('subprocess.Popen')
    def test_rename_containers(self, popen):
        ps_result = '''one one
two-12345678 two
two two
three-12345678 three
four-12345678 four
'''

        self.mock_execute(popen, ps_result, '', 0)
        self.runner.rename_container = mock.Mock()

        self.runner.rename_containers()

        self.assert_execute(
            popen, ['docker', 'ps', '-a',
                    '--filter', 'label=managed_by=tester',
                    '--format', '{{.Names}} {{.Label "container_name"}}']
        )
        # only containers three-12345678 and four-12345678 four will be renamed
        self.runner.rename_container.assert_has_calls([
            mock.call('three-12345678', 'three'),
            mock.call('four-12345678', 'four')
        ], any_order=True)


class PodmanRunner(TestBaseRunner):

    @mock.patch('subprocess.Popen')
    def test_image_exist(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.runner = runner.PodmanRunner('tester')
        self.runner.image_exist('one')
        self.assert_execute(
            popen, ['podman', 'image', 'exists', 'one']
        )

    @mock.patch('subprocess.Popen')
    def test_container_exist(self, popen):
        self.mock_execute(popen, '', '', 0)

        self.runner = runner.PodmanRunner('tester')
        self.runner.container_exist('one')
        self.assert_execute(
            popen, ['podman', 'container', 'exists', 'one']
        )

    @mock.patch('paunch.runner.BaseRunner.stop_container')
    def test_stop_container(self, mock_stop):
        mock_running = mock.MagicMock()
        mock_running.return_value = True
        self.runner = runner.PodmanRunner('tester')
        self.runner.container_running = mock_running
        self.runner.stop_container('foo')
        mock_running.assert_called_once_with('foo')
        mock_stop.assert_called_once_with('foo', None, False)

    @mock.patch('paunch.runner.BaseRunner.stop_container')
    def test_stop_container_missing(self, mock_stop):
        mock_running = mock.MagicMock()
        mock_running.return_value = False
        self.runner = runner.PodmanRunner('tester')
        self.runner.container_running = mock_running
        self.runner.stop_container('foo')
        mock_running.assert_called_once_with('foo')
        mock_stop.assert_not_called()

    @mock.patch('paunch.utils.systemd.service_delete')
    @mock.patch('paunch.runner.BaseRunner.remove_container')
    def test_remove_container(self, mock_remove, mock_systemd):
        mock_exists = mock.MagicMock()
        mock_exists.return_value = True
        self.runner = runner.PodmanRunner('tester')
        self.runner.container_exist = mock_exists
        self.runner.remove_container('foo')
        mock_exists.assert_called_once_with('foo')
        mock_systemd.assert_called_once_with(container='foo',
                                             log=self.runner.log)
        mock_remove.assert_called_once_with('foo')

    @mock.patch('paunch.utils.systemd.service_delete')
    @mock.patch('paunch.runner.BaseRunner.remove_container')
    def test_remove_container_missing(self, mock_remove, mock_systemd):
        mock_exists = mock.MagicMock()
        mock_exists.return_value = False
        self.runner = runner.PodmanRunner('tester')
        self.runner.container_exist = mock_exists
        self.runner.remove_container('foo')
        mock_exists.assert_called_once_with('foo')
        mock_systemd.assert_not_called()
        mock_remove.assert_not_called()



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import logging
import os
import psutil
import re
import sys
import yaml

from paunch import constants
from paunch import utils


def configure_logging(name, level=3, log_file=None):
    '''Mimic oslo_log default levels and formatting for the logger. '''
    log = logging.getLogger(name)

    if level and level > 2:
        ll = logging.DEBUG
    elif level and level == 2:
        ll = logging.INFO
    else:
        ll = logging.WARNING

    log.setLevel(ll)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(ll)
    if log_file:
        fhandler = logging.FileHandler(log_file)
        formatter = logging.Formatter(
            '%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
            '%(name)s [  ] %(message)s',
            '%Y-%m-%d %H:%M:%S')
        fhandler.setLevel(ll)
        fhandler.setFormatter(formatter)
        log.addHandler(fhandler)
        log.addHandler(handler)
        log.propagate = False

    return log


def configure_logging_from_args(name, app_args):
    # takes 1, or 2 if --verbose, or 4 - 5 if --debug
    log_level = (app_args.verbose_level +
                 int(app_args.debug) * 3)

    # if executed as root log to specified file or default log file
    if os.getuid() == 0:
        log_file = app_args.log_file or constants.LOG_FILE
    else:
        log_file = app_args.log_file

    log = utils.common.configure_logging(
        __name__, log_level, log_file)
    return (log, log_file, log_level)


def get_cpus_allowed_list(**args):
    """Returns the process's Cpus_allowed on which CPUs may be scheduled.

    :return: Value for Cpus_allowed, e.g. '0-3'
    """
    return ','.join([str(c) for c in psutil.Process().cpu_affinity()])


def load_config(config, name=None, overrides=None):
    container_config = {}
    if overrides is None:
        overrides = {}
    if os.path.isdir(config):
        # When the user gives a config directory and specify a container name,
        # we return the container config for that specific container.
        if name:
            cf = 'hashed-' + name + '.json'
            with open(os.path.join(config, cf), 'r') as f:
                container_config[name] = {}
                container_config[name].update(yaml.safe_load(f))
        # When the user gives a config directory and without container name,
        # we return all container configs in that directory.
        else:
            config_files = glob.glob(os.path.join(config, 'hashed-*.json'))
            for cf in config_files:
                with open(os.path.join(config, cf), 'r') as f:
                    name = os.path.basename(os.path.splitext(
                        cf.replace('hashed-', ''))[0])
                    container_config[name] = {}
                    container_config[name].update(yaml.safe_load(f))
    else:
        # Backward compatibility so our users can still use the old path,
        # paunch will recognize it and find the right container config.
        old_format = '/var/lib/tripleo-config/hashed-container-startup-config'
        if config.startswith(old_format):
            step = re.search('/var/lib/tripleo-config/'
                             'hashed-container-startup-config-step'
                             '_(.+).json', config).group(1)
            # If a name is specified, we return the container config for that
            # specific container.
            if name:
                new_path = os.path.join(
                    '/var/lib/tripleo-config/container_startup_config',
                    'step_' + step, 'hashed-' + name + '.json')
                with open(new_path, 'r') as f:
                    c_config = yaml.safe_load(f)
                    container_config[name] = {}
                    container_config[name].update(c_config[name])
            # When no name is specified, we return all container configs in
            # the file.
            else:
                new_path = os.path.join(
                    '/var/lib/tripleo-config/container_startup_config',
                    'step_' + step)
                config_files = glob.glob(os.path.join(new_path,
                                                      'hashed-*.json'))
                for cf in config_files:
                    with open(os.path.join(new_path, cf), 'r') as f:
                        name = os.path.basename(os.path.splitext(
                            cf.replace('hashed-', ''))[0])
                        c_config = yaml.safe_load(f)
                        container_config[name] = {}
                        container_config[name].update(c_config[name])
        # When the user gives a file path, that isn't the old format,
        # we consider it's the new format so the file name is the container
        # name.
        else:
            if not name:
                # No name was given, we'll guess it with file name
                name = os.path.basename(os.path.splitext(
                    config.replace('hashed-', ''))[0])
            with open(os.path.join(config), 'r') as f:
                container_config[name] = {}
                container_config[name].update(yaml.safe_load(f))

    # Overrides
    for k in overrides.keys():
        if k in container_config:
            for mk, mv in overrides[k].items():
                container_config[k][mk] = mv

    return container_config



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from paunch.tests import base
from paunch.utils import systemctl


class TestUtilsSystemctl(base.TestCase):

    def test_format_service_name(self):
        expected = 'test.service'
        self.assertEqual(expected, systemctl.format_name('test'))
        self.assertEqual(expected, systemctl.format_name(expected))

    @mock.patch('subprocess.check_call', autospec=True)
    def test_stop(self, mock_subprocess_check_call):
        test = 'test'
        systemctl.stop(test)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'stop', test]),
        ])

    @mock.patch('subprocess.check_call', autospec=True)
    def test_daemon_reload(self, mock_subprocess_check_call):
        systemctl.daemon_reload()
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'daemon-reload']),
        ])

    @mock.patch('subprocess.check_call', autospec=True)
    def test_is_active(self, mock_subprocess_check_call):
        systemctl.is_active('foo')
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'is-active', '-q', 'foo']),
        ])

    @mock.patch('subprocess.check_call', autospec=True)
    def test_enable(self, mock_subprocess_check_call):
        test = 'test'
        systemctl.enable(test, now=True)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'enable', '--now', test]),
        ])
        systemctl.enable(test)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'enable', '--now', test]),
        ])
        systemctl.enable(test, now=False)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'enable', test]),
        ])

    @mock.patch('subprocess.check_call', autospec=True)
    def test_disable(self, mock_subprocess_check_call):
        test = 'test'
        systemctl.disable(test)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'disable', test]),
        ])

    @mock.patch('subprocess.check_call', autospec=True)
    def test_add_requires(self, mock_subprocess_check_call):
        test = 'test'
        requires = "foo"
        systemctl.add_requires(test, requires)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'add-requires', test, requires]),
        ])
        requires = ["foo", "bar"]
        systemctl.add_requires(test, requires)
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'add-requires', test, "foo", "bar"]),
        ])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import distutils.spawn
import itertools
import json
import os
import re
import shutil
import tenacity
import yaml

from paunch.utils import common
from paunch.utils import systemd


class BaseBuilder(object):

    def __init__(self, config_id, config, runner, labels, log=None,
                 cont_log_path=None, healthcheck_disabled=False, cleanup=True):
        self.config_id = config_id
        self.config = config
        self.labels = labels
        self.runner = runner
        # Leverage pre-configured logger
        self.log = log or common.configure_logging(__name__)
        self.cont_log_path = cont_log_path
        self.healthcheck_disabled = healthcheck_disabled
        self.cleanup = cleanup

        if os.path.isfile('/var/lib/tripleo-config/.ansible-managed'):
            msg = ('Containers were previously deployed with '
                   'tripleo-ansible, paunch CLI can not be used.')
            raise RuntimeError(msg)

    def apply(self):

        stdout = []
        stderr = []
        pull_returncode = self.pull_missing_images(stdout, stderr)
        if pull_returncode != 0:
            return stdout, stderr, pull_returncode

        deploy_status_code = 0
        key_fltr = lambda k: self.config[k].get('start_order', 0)

        failed_containers = []
        container_names = self.runner.container_names(self.config_id)

        # Cleanup containers missing from the config.
        # Also applying new containers configs is an opportunity for
        # renames to their preferred names.
        changed = self.delete_missing(container_names)
        changed |= self.runner.rename_containers()
        if changed:
            # If anything has been changed, refresh the container_names
            container_names = self.runner.container_names(self.config_id)
        desired_names = set([cn[-1] for cn in container_names])

        for container in sorted(self.config, key=key_fltr):
            # Before creating the container, figure out if it needs to be
            # removed because of its configuration has changed.
            # If anything has been deleted, refresh the container_names/desired
            if self.delete_updated(container, container_names):
                container_names = self.runner.container_names(self.config_id)
                desired_names = set([cn[-1] for cn in container_names])

            self.log.debug("Running container: %s" % container)
            cconfig = self.config[container]
            action = cconfig.get('action', 'run')
            restart = cconfig.get('restart', 'none')
            exit_codes = cconfig.get('exit_codes', [0])
            container_name = self.runner.unique_container_name(container)
            systemd_managed = (restart != 'none'
                               and self.runner.cont_cmd == 'podman'
                               and action == 'run')
            start_cmd = 'create' if systemd_managed else 'run'

            # When upgrading from Docker to Podman, we want to stop the
            # container that runs under Docker first before starting it with
            # Podman. The container will be removed later in THT during
            # upgrade_tasks.
            if self.runner.cont_cmd == 'podman' and \
                    os.path.exists('/var/run/docker.sock'):
                self.runner.stop_container(container, 'docker', quiet=True)

            if action == 'run':
                if container in desired_names:
                    self.log.debug('Skipping existing container: %s' %
                                   container)
                    continue

                c_name = self.runner.discover_container_name(
                    container, self.config_id) or container
                cmd = [
                    self.runner.cont_cmd,
                    start_cmd,
                    '--name',
                    c_name
                ]

                self.label_arguments(cmd, container)
                self.log.debug("Start container {} as {}.".format(container,
                                                                  c_name))
                validations_passed = self.container_run_args(
                    cmd, container, c_name)
            elif action == 'exec':
                # for exec, the first argument is the fixed named container
                # used when running the command into the running container.
                # use the discovered container name to manipulate with the
                # real (delagate) container representing the fixed named one
                command = self.command_argument(cconfig.get('command'))
                if command:
                    c_name = self.runner.discover_container_name(
                        command[0], self.config_id)
                else:
                    c_name = self.runner.discover_container_name(
                        container, self.config_id)
                # Before running the exec, we want to make sure the container
                # is running.
                # https://bugs.launchpad.net/bugs/1839559
                if not c_name or not self.runner.container_running(c_name):
                    msg = ('Failing to apply action exec for '
                           'container: %s' % container)
                    raise RuntimeError(msg)
                cmd = [self.runner.cont_cmd, 'exec']
                validations_passed = self.cont_exec_args(cmd,
                                                         container,
                                                         c_name)

            if not validations_passed:
                self.log.debug('Validations failed. Skipping container: %s' %
                               container)
                failed_containers.append(container)
                continue

            (cmd_stdout, cmd_stderr, returncode) = self.runner.execute(
                cmd, self.log)
            if cmd_stdout:
                stdout.append(cmd_stdout)
            if cmd_stderr:
                stderr.append(cmd_stderr)

            if returncode not in exit_codes:
                self.log.error("Error running %s. [%s]\n" % (cmd, returncode))
                self.log.error("stdout: %s" % cmd_stdout)
                self.log.error("stderr: %s" % cmd_stderr)
                deploy_status_code = returncode
            else:
                self.log.debug('Completed $ %s' % ' '.join(cmd))
                self.log.info("stdout: %s" % cmd_stdout)
                self.log.info("stderr: %s" % cmd_stderr)
                if systemd_managed:
                    systemd.service_create(container=container_name,
                                           cconfig=cconfig,
                                           log=self.log)
                    if (not self.healthcheck_disabled and
                            'healthcheck' in cconfig):
                        check = cconfig.get('healthcheck')['test']
                        systemd.healthcheck_create(container=container_name,
                                                   log=self.log, test=check)
                        systemd.healthcheck_timer_create(
                            container=container_name,
                            cconfig=cconfig,
                            log=self.log)

        if failed_containers:
            message = (
                "The following containers failed validations "
                "and were not started: {}".format(
                    ', '.join(failed_containers)))
            self.log.error(message)
            # The message is also added to stderr so that it's returned and
            # logged by the paunch module for ansible
            stderr.append(message)
            deploy_status_code = 1

        return stdout, stderr, deploy_status_code

    def delete_missing(self, container_names):
        deleted = False
        for cn in container_names:
            container = cn[0]
            # if the desired name is not in the config, delete it
            if cn[-1] not in self.config:
                if self.cleanup:
                    self.log.debug("Deleting container (removed): "
                                   "%s" % container)
                    self.runner.remove_container(container)
                    deleted = True
                else:
                    self.log.debug("Skipping container (cleanup disabled): "
                                   "%s" % container)
        return deleted

    def delete_updated(self, container, container_names):
        # If a container is not deployed, there is nothing to delete
        if (container not in
           list(itertools.chain.from_iterable(container_names))):
            return False

        # fetch container inspect info
        inspect_info = self.runner.inspect(container)
        if not inspect_info:
            # we shouldn't get here but you never know
            self.log.debug("Deleting container (no inspect data): "
                           "%s" % container)
            self.runner.remove_container(container)
            return True
        container_config = inspect_info.get('Config', {})
        config_data = container_config.get('Labels', {}).get('config_data')
        try:
            ex_data = yaml.safe_load(str(config_data))
        except Exception:
            ex_data = None

        if not ex_data:
            if self.cleanup:
                self.log.debug("Deleting container (no_config_data): "
                               "%s" % container)
                self.runner.remove_container(container)
                return True
            else:
                self.log.debug("Skipping container (cleanup disabled): "
                               "%s" % container)

        # check if config_data has changed
        new_data = self.config[container]
        if new_data != ex_data:
            self.log.debug("Deleting container (changed config_data): %s"
                           % container)
            self.runner.remove_container(container)
            return True

        # check if the container image has changed (but tag as not)
        # e.g. if you use :latest, the name doesn't change but the ID does
        container_image = container_config.get('Image')

        if container_image:
            image_id_str = self.runner.inspect(
                container_image, "{{.Id}}", o_type='image')
            if str(image_id_str).strip() != inspect_info.get('Image'):
                self.log.debug("Deleting container (image updated): "
                               "%s" % container)
                self.runner.remove_container(container)
                return True

        return False

    def label_arguments(self, cmd, container):
        if self.labels:
            for i, v in self.labels.items():
                cmd.extend(['--label', '%s=%s' % (i, v)])
        cmd.extend([
            '--label',
            'config_id=%s' % self.config_id,
            '--label',
            'container_name=%s' % container,
            '--label',
            'managed_by=%s' % self.runner.managed_by,
            '--label',
            'config_data=%s' % json.dumps(self.config.get(container))
        ])

    def boolean_arg(self, cconfig, cmd, key, arg):
        if cconfig.get(key, False):
            cmd.append(arg)

    def string_arg(self, cconfig, cmd, key, arg, transform=None):
        if key in cconfig:
            if transform:
                value = transform(cconfig[key])
            else:
                value = cconfig[key]
            cmd.append('%s=%s' % (arg, value))

    def list_or_string_arg(self, cconfig, cmd, key, arg):
        if key not in cconfig:
            return
        value = cconfig[key]
        if not isinstance(value, list):
            value = [value]
        for v in value:
            if v:
                cmd.append('%s=%s' % (arg, v))

    def list_arg(self, cconfig, cmd, key, arg):
        if key not in cconfig:
            return
        value = cconfig[key]
        for v in value:
            if v:
                cmd.append('%s=%s' % (arg, v))

    def list_or_dict_arg(self, cconfig, cmd, key, arg):
        if key not in cconfig:
            return
        value = cconfig[key]
        if isinstance(value, dict):
            for k, v in sorted(value.items()):
                if v:
                    cmd.append('%s=%s=%s' % (arg, k, v))
                elif k:
                    cmd.append('%s=%s' % (arg, k))
        elif isinstance(value, list):
            for v in value:
                if v:
                    cmd.append('%s=%s' % (arg, v))

    def cont_exec_args(self, cmd, container, delegate=None):
        """Prepare the exec command args, from the container configuration.

        :param cmd: The list of command options to be modified
        :param container: A dict with container configurations
        :param delegate: A predictable/unique name of the actual container
        :returns: True if configuration is valid, otherwise False
        """
        if delegate and container != delegate:
            self.log.debug("Container {} has a delegate "
                           "{}".format(container, delegate))
        cconfig = self.config[container]
        if 'privileged' in cconfig:
            cmd.append('--privileged=%s' % str(cconfig['privileged']).lower())
        if 'user' in cconfig:
            cmd.append('--user=%s' % cconfig['user'])
        self.list_or_dict_arg(cconfig, cmd, 'environment', '--env')
        command = self.command_argument(cconfig.get('command'))
        # for exec, the first argument is the container name,
        # make sure the correct one is used
        if command:
            if not delegate:
                command[0] = self.runner.discover_container_name(
                    command[0], self.config_id)
            else:
                command[0] = delegate
        cmd.extend(command)

        return True

    def pull_missing_images(self, stdout, stderr):
        images = set()
        for container in self.config:
            cconfig = self.config[container]
            image = cconfig.get('image')
            if image:
                images.add(image)

        returncode = 0

        for image in sorted(images):

            # only pull if the image does not exist locally
            if self.runner.cont_cmd == 'docker':
                if self.runner.inspect(image,
                                       output_format='exists',
                                       o_type='image'):
                    continue
            else:
                img_exist = self.runner.image_exist(image)
                if img_exist:
                    continue

            try:
                (cmd_stdout, cmd_stderr) = self._pull(image)
            except PullException as e:
                returncode = e.rc
                cmd_stdout = e.stdout
                cmd_stderr = e.stderr
                self.log.error("Error pulling %s. [%s]\n" %
                               (image, returncode))
                self.log.error("stdout: %s" % e.stdout)
                self.log.error("stderr: %s" % e.stderr)
            else:
                self.log.debug('Pulled %s' % image)
                self.log.info("stdout: %s" % cmd_stdout)
                self.log.info("stderr: %s" % cmd_stderr)

            if cmd_stdout:
                stdout.append(cmd_stdout)
            if cmd_stderr:
                stderr.append(cmd_stderr)

        return returncode

    @tenacity.retry(  # Retry up to 4 times with jittered exponential backoff
        reraise=True,
        wait=tenacity.wait_random_exponential(multiplier=1, max=10),
        stop=tenacity.stop_after_attempt(4)
    )
    def _pull(self, image):
        cmd = [self.runner.cont_cmd, 'pull', image]
        (stdout, stderr, rc) = self.runner.execute(cmd, self.log)
        if rc != 0:
            raise PullException(stdout, stderr, rc)
        return stdout, stderr

    @staticmethod
    def command_argument(command):
        if not command:
            return []
        if not isinstance(command, list):
            return command.split()
        return command

    def lower(self, a):
        return str(a).lower()

    def which(self, program):
        try:
            pgm = shutil.which(program)
        except AttributeError:
            pgm = distutils.spawn.find_executable(program)
        return pgm

    def duration(self, a):
        if isinstance(a, (int, float)):
            return a

        # match groups of the format 5h34m56s
        m = re.match('^'
                     '(([\d\.]+)h)?'
                     '(([\d\.]+)m)?'
                     '(([\d\.]+)s)?'
                     '(([\d\.]+)ms)?'
                     '(([\d\.]+)us)?'
                     '$', a)

        if not m:
            # fallback to parsing string as a number
            return float(a)

        n = 0.0
        if m.group(2):
            n += 3600 * float(m.group(2))
        if m.group(4):
            n += 60 * float(m.group(4))
        if m.group(6):
            n += float(m.group(6))
        if m.group(8):
            n += float(m.group(8)) / 1000.0
        if m.group(10):
            n += float(m.group(10)) / 1000000.0
        return n

    def validate_volumes(self, volumes):
        """Validate volume sources

        Validates that the source volume either exists on the filesystem
        or is a valid container volume.  Since podman will error if the
        source volume filesystem path doesn't exist, we want to catch the
        error before podman.

        :param: volumes: list of volume mounts in the format of "src:path"
        """
        valid = True
        for volume in volumes:
            if not volume:
                # ignore when volume is ''
                continue
            src_path = volume.split(':', 1)[0]
            check = self.runner.validate_volume_source(src_path)
            if not check:
                self.log.error("%s is not a valid volume source" % src_path)
                valid = False
        return valid


class PullException(Exception):

    def __init__(self, stdout, stderr, rc):
        self.stdout = stdout
        self.stderr = stderr
        self.rc = rc



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import tempfile

from paunch.tests import base
from paunch.utils import systemd


class TestUtilsSystemd(base.TestCase):

    @mock.patch('shutil.rmtree', autospec=True)
    @mock.patch('os.path.exists', autospec=True)
    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('os.chmod')
    def test_service_create(self, mock_chmod, mock_subprocess_check_call,
                            mock_exists, mock_rmtree):
        container = 'my_app'
        service = 'tripleo_' + container
        cconfig = {'depends_on': ['something'], 'restart': 'unless-stopped',
                   'stop_grace_period': '15'}
        tempdir = tempfile.mkdtemp()
        systemd.service_create(container, cconfig, tempdir)

        sysd_unit_f = tempdir + service + '.service'
        unit = open(sysd_unit_f, 'rt').read()
        self.assertIn('Wants=something.service', unit)
        self.assertIn('Restart=always', unit)
        self.assertIn('ExecStop=/usr/bin/podman stop -t 15 my_app', unit)
        self.assertIn('ExecStopPost=/usr/bin/podman stop -t 15 my_app', unit)
        self.assertIn('PIDFile=/var/run/my_app.pid', unit)
        mock_chmod.assert_has_calls([mock.call(sysd_unit_f, 420)])

        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'daemon-reload']),
            mock.call(['systemctl', 'enable', '--now', service]),
        ])

        mock_rmtree.assert_has_calls([
            mock.call(sysd_unit_f + '.requires')
        ])

        os.rmdir(tempdir)

    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('os.chmod')
    def test_svc_extended_create(self, mock_chmod, mock_subprocess_check_call):
        container = 'my_app'
        service = 'tripleo_' + container
        cconfig = {'depends_on': ['something'], 'restart': 'unless-stopped',
                   'stop_grace_period': '15',
                   'systemd_exec_flags': {'RootDirectory': '/srv',
                                          'LimitCPU': '60',
                                          'RuntimeDirectory': 'my_app foo/bar'}
                   }
        tempdir = tempfile.mkdtemp()
        systemd.service_create(container, cconfig, tempdir)

        sysd_unit_f = tempdir + service + '.service'
        unit = open(sysd_unit_f, 'rt').read()
        self.assertIn('RootDirectory=/srv', unit)
        self.assertIn('LimitCPU=60', unit)
        self.assertIn('RuntimeDirectory=my_app foo/bar', unit)

        os.rmdir(tempdir)

    @mock.patch('shutil.rmtree', autospec=True)
    @mock.patch('os.remove', autospec=True)
    @mock.patch('os.path.exists', autospec=True)
    @mock.patch('os.path.isfile', autospec=True)
    @mock.patch('subprocess.check_call', autospec=True)
    def test_service_delete(self, mock_subprocess_check_call, mock_isfile,
                            mock_exists, mock_rm, mock_rmtree):
        mock_isfile.return_value = True
        container = 'my_app'
        service = 'tripleo_' + container
        tempdir = tempfile.mkdtemp()
        service_requires_d = service + '.service.requires'

        systemd.service_delete(container, tempdir)
        mock_rm.assert_has_calls([
            mock.call(tempdir + service + '.service'),
            mock.call(tempdir + service + '_healthcheck.service'),
            mock.call(tempdir + service + '_healthcheck.timer'),
        ])
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'stop', service + '.service']),
            mock.call(['systemctl', 'disable', service + '.service']),
            mock.call(['systemctl', 'stop', service + '_healthcheck.service']),
            mock.call(['systemctl', 'disable', service +
                       '_healthcheck.service']),
            mock.call(['systemctl', 'stop', service + '_healthcheck.timer']),
            mock.call(['systemctl', 'disable', service +
                       '_healthcheck.timer']),
        ])
        mock_rmtree.assert_has_calls([
            mock.call(os.path.join(tempdir, service_requires_d)),
        ])

    @mock.patch('os.chmod')
    def test_healthcheck_create(self, mock_chmod):
        container = 'my_app'
        service = 'tripleo_' + container
        tempdir = tempfile.mkdtemp()
        healthcheck = service + '_healthcheck.service'
        sysd_unit_f = tempdir + healthcheck

        systemd.healthcheck_create(container, tempdir)
        unit = open(sysd_unit_f, 'rt').read()

        self.assertIn('Requisite=tripleo_my_app.service', unit)
        self.assertIn('ExecStart=/usr/bin/podman exec --user root my_app '
                      '/openstack/healthcheck', unit)
        mock_chmod.assert_has_calls([mock.call(sysd_unit_f, 420)])

    @mock.patch('os.chmod')
    def test_healthcheck_create_command(self, mock_chmod):
        container = 'my_app'
        service = 'tripleo_' + container
        tempdir = tempfile.mkdtemp()
        healthcheck = service + '_healthcheck.service'
        sysd_unit_f = tempdir + healthcheck
        check = '/foo/bar baz'

        systemd.healthcheck_create(container, tempdir, test=check)
        unit = open(sysd_unit_f, 'rt').read()
        self.assertIn('ExecStart=/usr/bin/podman exec --user root my_app '
                      '/foo/bar baz', unit)

    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('os.chmod')
    def test_healthcheck_timer_create(self, mock_chmod,
                                      mock_subprocess_check_call):
        container = 'my_app'
        service = 'tripleo_' + container
        cconfig = {'check_interval': '15'}
        tempdir = tempfile.mkdtemp()
        healthcheck_timer = service + '_healthcheck.timer'
        sysd_unit_f = tempdir + healthcheck_timer

        systemd.healthcheck_timer_create(container, cconfig, tempdir)
        unit = open(sysd_unit_f, 'rt').read()

        self.assertIn('PartOf=%s.service' % service, unit)
        self.assertIn('OnActiveSec=120', unit)
        self.assertIn('OnUnitActiveSec=15', unit)
        mock_chmod.assert_has_calls([mock.call(sysd_unit_f, 420)])
        mock_subprocess_check_call.assert_has_calls([
            mock.call(['systemctl', 'enable', '--now', healthcheck_timer]),
            mock.call(['systemctl', 'add-requires', service + '.service',
                      healthcheck_timer]),
            mock.call(['systemctl', 'daemon-reload']),
        ])



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import subprocess
import tenacity

from paunch.utils import common


class SystemctlException(Exception):
    pass


def systemctl(cmd, log=None):
    log = log or common.configure_logging(__name__)
    if not isinstance(cmd, list):
        raise SystemctlException("systemctl cmd passed must be a list")
    cmd.insert(0, 'systemctl')
    log.debug("Executing: {}".format(" ".join(cmd)))
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as err:
        raise SystemctlException(str(err))


def format_name(name):
    return name if name.endswith('.service') else name + ".service"


def stop(service, log=None):
    systemctl(['stop', service], log)


def daemon_reload(log=None):
    systemctl(['daemon-reload'], log)


def reset_failed(service, log=None):
    systemctl(['reset-failed', service], log)


def is_active(service, log=None):
    systemctl(['is-active', '-q', service], log)


# NOTE(bogdando): this implements a crash-loop with reset-failed
# counters approach that provides an efficient feature parity to the
# classic rate limiting, shall we want to implement that for the
# systemctl command wrapper instead.
@tenacity.retry(  # Retry up to 5 times with jittered exponential backoff
    reraise=True,
    retry=tenacity.retry_if_exception_type(
        SystemctlException
    ),
    wait=tenacity.wait_random_exponential(multiplier=1, max=10),
    stop=tenacity.stop_after_attempt(5)
)
def enable(service, now=True, log=None):
    cmd = ['enable']
    if now:
        cmd.append('--now')
    cmd.append(service)
    try:
        systemctl(cmd, log)
    except SystemctlException as err:
        # Reset failure counters for the service unit and retry
        reset_failed(service, log)
        raise SystemctlException(str(err))


def disable(service, log=None):
    systemctl(['disable', service], log)


def add_requires(target, units, log=None):
    cmd = ['add-requires', target]
    if isinstance(units, list):
        cmd.extend(units)
    else:
        cmd.append(units)
    systemctl(cmd, log)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from paunch.builder import base
from paunch.utils import common


class ComposeV1Builder(base.BaseBuilder):

    def __init__(self, config_id, config, runner, labels=None, log=None,
                 cleanup=True):
        super(ComposeV1Builder, self).__init__(config_id, config, runner,
                                               labels, log, cleanup)

    def container_run_args(self, cmd, container, delegate=None):
        """Prepare the run command args, from the container configuration.

        :param cmd: The list of command options to be modified
        :param container: A dict with container configurations
        :delegate: A compatibility parameter for podman, does nothing here
        :returns: True if configuration is valid, otherwise False
        """
        if delegate and container != delegate:
            self.log.debug("Delegate {} of container {} has no special "
                           "meanings for this context and will be "
                           "ignored".format(delegate, container))
        cconfig = self.config[container]
        if cconfig.get('detach', True):
            cmd.append('--detach=true')
        self.list_or_string_arg(cconfig, cmd, 'env_file', '--env-file')
        self.list_or_dict_arg(cconfig, cmd, 'environment', '--env')
        self.boolean_arg(cconfig, cmd, 'remove', '--rm')
        self.boolean_arg(cconfig, cmd, 'interactive', '--interactive')
        self.boolean_arg(cconfig, cmd, 'tty', '--tty')
        self.string_arg(cconfig, cmd, 'net', '--net')
        self.string_arg(cconfig, cmd, 'ipc', '--ipc')
        self.string_arg(cconfig, cmd, 'pid', '--pid')
        self.string_arg(cconfig, cmd, 'uts', '--uts')
        # TODO(sbaker): implement ulimits property, deprecate this ulimit
        # property
        for u in cconfig.get('ulimit', []):
            if u:
                cmd.append('--ulimit=%s' % u)
        if 'healthcheck' in cconfig:
            hconfig = cconfig['healthcheck']
            if 'test' in hconfig:
                cmd.append('--health-cmd')
                cmd.append(str(hconfig['test']))
            if 'interval' in hconfig:
                cmd.append('--health-interval=%s' % hconfig['interval'])
            if 'timeout' in hconfig:
                cmd.append('--health-timeout=%s' % hconfig['timeout'])
            if 'retries' in hconfig:
                cmd.append('--health-retries=%s' % hconfig['retries'])

        self.string_arg(cconfig, cmd, 'privileged', '--privileged', self.lower)
        self.string_arg(cconfig, cmd, 'restart', '--restart')
        self.string_arg(cconfig, cmd, 'user', '--user')
        self.list_arg(cconfig, cmd, 'group_add', '--group-add')
        self.list_arg(cconfig, cmd, 'volumes', '--volume')
        self.list_arg(cconfig, cmd, 'volumes_from', '--volumes-from')
        # TODO(sbaker): deprecate log_tag, implement log_driver, log_opt
        if 'log_tag' in cconfig:
            cmd.append('--log-opt=tag=%s' % cconfig['log_tag'])
        self.string_arg(cconfig, cmd, 'cpu_shares', '--cpu-shares')
        self.string_arg(cconfig, cmd, 'mem_limit', '--memory')
        self.string_arg(cconfig, cmd, 'memswap_limit', '--memory-swap')
        self.string_arg(cconfig, cmd, 'mem_swappiness', '--memory-swappiness')
        self.list_or_string_arg(cconfig, cmd, 'security_opt', '--security-opt')
        self.string_arg(cconfig, cmd, 'stop_signal', '--stop-signal')
        self.string_arg(cconfig, cmd, 'hostname', '--hostname')
        for extra_host in cconfig.get('extra_hosts', []):
            if extra_host:
                cmd.append('--add-host=%s' % extra_host)
        if 'cpuset_cpus' in cconfig:
            # 'all' is a special value to directly configure all CPUs
            # that are available. Without specifying --cpuset-cpus, we'll
            # let the container engine to figure out what CPUs are online.
            # https://bugs.launchpad.net/tripleo/+bug/1868135
            # https://bugzilla.redhat.com/show_bug.cgi?id=1813091
            if cconfig['cpuset_cpus'] != 'all':
                cmd.append('--cpuset-cpus=%s' % cconfig['cpuset_cpus'])
        else:
            with open('/proc/cmdline') as cmdline:
                if 'isolcpus' in cmdline.read():
                    cmd.append('--cpuset-cpus=%s' %
                               common.get_cpus_allowed_list())

        self.string_arg(cconfig, cmd,
                        'stop_grace_period', '--stop-timeout',
                        self.duration)

        self.list_arg(cconfig, cmd, 'cap_add', '--cap-add')
        self.list_arg(cconfig, cmd, 'cap_drop', '--cap-drop')

        # TODO(sbaker): add missing compose v1 properties:
        # cgroup_parent
        # devices
        # dns, dns_search
        # entrypoint
        # expose
        # extra_hosts
        # labels
        # ports
        # stop_signal
        # volume_driver
        # cpu_quota
        # domainname
        # hostname
        # mac_address
        # memory_reservation
        # kernel_memory
        # read_only
        # shm_size
        # stdin_open
        # working_dir

        cmd.append(cconfig.get('image', ''))
        cmd.extend(self.command_argument(cconfig.get('command')))

        return self.validate_volumes(cconfig.get('volumes', []))



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import inspect
import json
import mock
import sys
import tenacity

from paunch.builder import base as basebuilder
from paunch.builder import compose1
from paunch import runner
from paunch.tests import base


class TestBaseBuilder(base.TestCase):

    def setUp(self):
        super(TestBaseBuilder, self).setUp()
        mock_cpu_obj = mock.MagicMock()
        mock_cpu_aff = mock.MagicMock()
        mock_cpu_aff.return_value = [0, 1, 2, 3]
        mock_cpu_obj.cpu_affinity = mock_cpu_aff
        self.psutil_mock = mock.patch('psutil.Process',
                                      return_value=mock_cpu_obj)
        self.psutil_mock.start()
        self.addCleanup(self.psutil_mock.stop)

    @mock.patch("paunch.builder.base.BaseBuilder.delete_updated",
                return_value=False)
    def test_apply(self, mock_delete_updated):
        orig_call = tenacity.wait.wait_random_exponential.__call__
        orig_argspec = inspect.getargspec(orig_call)
        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            },
            'two': {
                'start_order': 1,
                'image': 'centos:7',
            },
            'three': {
                'start_order': 2,
                'image': 'centos:6',
            },
            'four': {
                'start_order': 10,
                'image': 'centos:7',
            },
            'four_ls': {
                'action': 'exec',
                'start_order': 20,
                'command': ['four', 'ls', '-l', '/']
            }
        }

        r = runner.DockerRunner(managed_by='tester', cont_cmd='docker')
        exe = mock.Mock()
        exe.side_effect = [
            ('exists', '', 0),  # inspect for image centos:6
            ('', '', 1),  # inspect for missing image centos:7
            ('Pulled centos:7', 'ouch', 1),  # pull centos:6 fails
            ('Pulled centos:7', '', 0),  # pull centos:6 succeeds
            # container_names for delete_missing (twice by managed_by)
            ('', '', 0),
            ('''five five
six six
two two
three-12345678 three''', '', 0),
            ('', '', 0),  # stop five
            ('', '', 0),  # rm five
            ('', '', 0),  # stop six
            ('', '', 0),  # rm six
            # container_names for rename_containers
            ('three-12345678 three', '', 0),
            ('', '', 0),  # rename three
            # desired/container_names to be refreshed after delete/rename
            ('three three', '', 0),  # renamed three already exists
            ('Created one-12345678', '', 0),
            ('Created two-12345678', '', 0),
            ('Created four-12345678', '', 0),
            ('a\nb\nc', '', 0)  # exec four
        ]
        r.discover_container_name = lambda n, c: '%s-12345678' % n
        r.unique_container_name = lambda n: '%s-12345678' % n
        r.execute = exe

        with mock.patch('tenacity.wait.wait_random_exponential.__call__') as f:
            f.return_value = 0
            with mock.patch('inspect.getargspec') as mock_args:
                mock_args.return_value = orig_argspec
                builder = compose1.ComposeV1Builder('foo', config, r)

        stdout, stderr, deploy_status_code = builder.apply()
        self.assertEqual(0, deploy_status_code)
        self.assertEqual([
            'Pulled centos:7',
            'Created one-12345678',
            'Created two-12345678',
            'Created four-12345678',
            'a\nb\nc'
        ], stdout)
        self.assertEqual([], stderr)

        exe.assert_has_calls([
            # inspect existing image centos:6
            mock.call(
                ['docker', 'inspect', '--type', 'image',
                 '--format', 'exists', 'centos:6'], mock.ANY, False, True
            ),
            # inspect and pull missing image centos:7
            mock.call(
                ['docker', 'inspect', '--type', 'image',
                 '--format', 'exists', 'centos:7'], mock.ANY, False, True
            ),
            # first pull attempt fails
            mock.call(
                ['docker', 'pull', 'centos:7'], mock.ANY
            ),
            # second pull attempt succeeds
            mock.call(
                ['docker', 'pull', 'centos:7'], mock.ANY
            ),
            # container_names for delete_missing
            mock.call(
                ['docker', 'ps', '-a',
                 '--filter', 'label=managed_by=tester',
                 '--filter', 'label=config_id=foo',
                 '--format', '{{.Names}} {{.Label "container_name"}}'],
                log=mock.ANY, quiet=False, warn_only=True
            ),
            mock.call(
                ['docker', 'ps', '-a',
                 '--filter', 'label=managed_by=paunch',
                 '--filter', 'label=config_id=foo',
                 '--format', '{{.Names}} {{.Label "container_name"}}'],
                mock.ANY
            ),
            # rm containers missing in config
            mock.call(['docker', 'stop', 'five'], mock.ANY),
            mock.call(['docker', 'rm', 'five'], mock.ANY),
            mock.call(['docker', 'stop', 'six'], mock.ANY),
            mock.call(['docker', 'rm', 'six'], mock.ANY),
            # container_names for rename
            mock.call(
                ['docker', 'ps', '-a',
                 '--filter', 'label=managed_by=tester',
                 '--format', '{{.Names}} {{.Label "container_name"}}'],
                log=mock.ANY, quiet=False, warn_only=True
            ),
            # rename three from an ephemeral to the static name
            mock.call(['docker', 'rename', 'three-12345678', 'three'],
                      mock.ANY),
            # container_names to be refreshed after delete/rename
            mock.call(
                ['docker', 'ps', '-a',
                 '--filter', 'label=managed_by=tester',
                 '--filter', 'label=config_id=foo',
                 '--format', '{{.Names}} {{.Label "container_name"}}'],
                log=mock.ANY, quiet=False, warn_only=True
            ),
            # run one
            mock.call(
                ['docker', 'run', '--name', 'one-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=one',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['one']),
                 '--detach=true',
                 'centos:7'], mock.ANY
            ),
            # run two
            mock.call(
                ['docker', 'run', '--name', 'two-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=two',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['two']),
                 '--detach=true',
                 'centos:7'], mock.ANY
            ),
            # run four
            mock.call(
                ['docker', 'run', '--name', 'four-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=four',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['four']),
                 '--detach=true',
                 'centos:7'], mock.ANY
            ),
            # execute within four
            mock.call(
                ['docker', 'exec', 'four-12345678', 'ls', '-l',
                 '/'], mock.ANY
            ),
        ])

    @mock.patch("paunch.runner.BaseRunner.container_names")
    @mock.patch("paunch.runner.BaseRunner.discover_container_name",
                return_value='one')
    def test_apply_idempotency_with_isolcpus(self, mock_dname, mock_cnames):
        config = {
            # running with the same config and given an ephemeral name
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            },
            # not running yet
            'two': {
                'start_order': 1,
                'image': 'centos:7',
            },
            # running, but with a different config
            'three': {
                'start_order': 2,
                'image': 'centos:7',
            },
            # not running yet
            'four': {
                'start_order': 10,
                'image': 'centos:7',
            },
            'one_ls': {
                'action': 'exec',
                'start_order': 20,
                'command': ['one', 'ls', '-l', '/']
            }
            # five is running but is not managed by us
        }
        # represents the state before and after renaming/removing things
        mock_cnames.side_effect = (
            # delete_missing
            [['five', 'five'], ['one-12345678', 'one'], ['three', 'three']],
            # rename_containers
            [['one-12345678', 'one']],
            # refresh container_names/desired after del/rename
            [['one', 'one'], ['three', 'three']],
            # refresh container_names/desired after delete_updated
            [['one', 'one']]
        )
        r = runner.DockerRunner(managed_by='tester', cont_cmd='docker')
        exe = mock.Mock()
        exe.side_effect = [
            # inspect for image centos:7
            ('exists', '', 0),
            # stop five
            ('', '', 0),
            # rm five
            ('', '', 0),
            ('', '', 0),  # ps for rename one
            # inspect one
            ('[{"Config": {"Labels": {"config_data": '
             '"{\\\"start_order\\\": 0, \\\"image\\\": \\\"centos:7\\\"}"'
             '}}}]', '', 0),
            ('Created two-12345678', '', 0),
            # inspect three
            ('[{"Config": {"Labels": {"config_data": '
             '"{\\\"start_order\\\": 42, \\\"image\\\": \\\"centos:7\\\"}"'
             '}}}]', '', 0),
            # stop three, changed config data
            ('', '', 0),
            # rm three, changed config data
            ('', '', 0),
            ('Created three-12345678', '', 0),
            ('Created four-12345678', '', 0),
            ('a\nb\nc', '', 0)  # exec one
        ]
        r.discover_container_name = lambda n, c: '%s-12345678' % n
        r.unique_container_name = lambda n: '%s-12345678' % n
        r.execute = exe

        builder = compose1.ComposeV1Builder('foo', config, r)
        if (sys.version_info > (3, 0)):
            open_builtins = "builtins.open"
        else:
            open_builtins = "__builtin__.open"
        with mock.patch(open_builtins,
                        mock.mock_open(read_data="isolcpus")):
            stdout, stderr, deploy_status_code = builder.apply()
        self.assertEqual(0, deploy_status_code)
        self.assertEqual([
            'Created two-12345678',
            'Created three-12345678',
            'Created four-12345678',
            'a\nb\nc'
        ], stdout)
        self.assertEqual([], stderr)

        exe.assert_has_calls([
            # inspect image centos:7
            mock.call(
                ['docker', 'inspect', '--type', 'image',
                 '--format', 'exists', 'centos:7'], mock.ANY, False, True
            ),
            # rm containers not in config
            mock.call(['docker', 'stop', 'five'], mock.ANY),
            mock.call(['docker', 'rm', 'five'], mock.ANY),
            # rename one from an ephemeral to the static name
            mock.call(['docker', 'rename', 'one-12345678', 'one'],
                      mock.ANY),
            # check the renamed one, config hasn't changed
            mock.call(['docker', 'inspect', '--type', 'container',
                       'one'], mock.ANY, False, True),
            # don't run one, its already running
            # run two
            mock.call(
                ['docker', 'run', '--name', 'two-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=two',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['two']),
                 '--detach=true', '--cpuset-cpus=0,1,2,3',
                 'centos:7'], mock.ANY
            ),
            # rm three, changed config
            mock.call(['docker', 'inspect', '--type', 'container',
                       'three'], mock.ANY, False, True),
            mock.call(['docker', 'stop', 'three'], mock.ANY),
            mock.call(['docker', 'rm', 'three'], mock.ANY),
            # run three
            mock.call(
                ['docker', 'run', '--name', 'three-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=three',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['three']),
                 '--detach=true', '--cpuset-cpus=0,1,2,3',
                 'centos:7'], mock.ANY
            ),
            # run four
            mock.call(
                ['docker', 'run', '--name', 'four-12345678',
                 '--label', 'config_id=foo',
                 '--label', 'container_name=four',
                 '--label', 'managed_by=tester',
                 '--label', 'config_data=%s' % json.dumps(config['four']),
                 '--detach=true', '--cpuset-cpus=0,1,2,3',
                 'centos:7'], mock.ANY
            ),
            # FIXME(bogdando): shall exec ls in the renamed one!
            # Why discover_container_name is never called to get it as c_name?
            mock.call(
                ['docker', 'exec', 'one-12345678', 'ls', '-l',
                 '/'], mock.ANY
            ),
        ])

    def test_apply_failed_pull(self):
        orig_call = tenacity.wait.wait_random_exponential.__call__
        orig_argspec = inspect.getargspec(orig_call)
        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            },
            'two': {
                'start_order': 1,
                'image': 'centos:7',
            },
            'three': {
                'start_order': 2,
                'image': 'centos:6',
            },
            'four': {
                'start_order': 10,
                'image': 'centos:7',
            },
            'four_ls': {
                'action': 'exec',
                'start_order': 20,
                'command': ['four', 'ls', '-l', '/']
            }
        }

        r = runner.DockerRunner(managed_by='tester', cont_cmd='docker')
        exe = mock.Mock()
        exe.side_effect = [
            ('exists', '', 0),  # inspect for image centos:6
            ('', '', 1),  # inspect for missing image centos:7
            ('Pulling centos:7', 'ouch', 1),  # pull centos:7 failure
            ('Pulling centos:7', 'ouch', 1),  # pull centos:7 retry 2
            ('Pulling centos:7', 'ouch', 1),  # pull centos:7 retry 3
            ('Pulling centos:7', 'ouch', 1),  # pull centos:7 retry 4
        ]
        r.execute = exe

        with mock.patch('tenacity.wait.wait_random_exponential.__call__') as f:
            f.return_value = 0
            with mock.patch('inspect.getargspec') as mock_args:
                mock_args.return_value = orig_argspec
                builder = compose1.ComposeV1Builder('foo', config, r)

        stdout, stderr, deploy_status_code = builder.apply()
        self.assertEqual(1, deploy_status_code)
        self.assertEqual(['Pulling centos:7'], stdout)
        self.assertEqual(['ouch'], stderr)

        exe.assert_has_calls([
            # inspect existing image centos:6
            mock.call(
                ['docker', 'inspect', '--type', 'image',
                 '--format', 'exists', 'centos:6'], mock.ANY, False, True
            ),
            # inspect and pull missing image centos:7
            mock.call(
                ['docker', 'inspect', '--type', 'image',
                 '--format', 'exists', 'centos:7'], mock.ANY, False, True
            ),
            mock.call(
                ['docker', 'pull', 'centos:7'], mock.ANY
            ),
        ])

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_label_arguments(self, runner):
        r = runner.return_value
        r.managed_by = 'tester'
        builder = compose1.ComposeV1Builder('foo', {}, r)
        cmd = []
        builder.label_arguments(cmd, 'one')
        self.assertEqual(
            ['--label', 'config_id=foo',
             '--label', 'container_name=one',
             '--label', 'managed_by=tester',
             '--label', 'config_data=null'],
            cmd)

        labels = collections.OrderedDict()
        labels['foo'] = 'bar'
        labels['bar'] = 'baz'

        builder = compose1.ComposeV1Builder('foo', {}, r, labels=labels)
        cmd = []
        builder.label_arguments(cmd, 'one')
        self.assertEqual(
            ['--label', 'foo=bar',
             '--label', 'bar=baz',
             '--label', 'config_id=foo',
             '--label', 'container_name=one',
             '--label', 'managed_by=tester',
             '--label', 'config_data=null'],
            cmd)

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_durations(self, runner):
        config = {
            'a': {'stop_grace_period': 123},
            'b': {'stop_grace_period': 123.5},
            'c': {'stop_grace_period': '123.3'},
            'd': {'stop_grace_period': '2.5s'},
            'e': {'stop_grace_period': '10s'},
            'f': {'stop_grace_period': '1m30s'},
            'g': {'stop_grace_period': '2h32m'},
            'h': {'stop_grace_period': '5h34m56s'},
            'i': {'stop_grace_period': '1h1m1s1ms1us'},
        }
        builder = compose1.ComposeV1Builder('foo', config, runner)

        result = {
            'a': '--stop-timeout=123',
            'b': '--stop-timeout=123.5',
            'c': '--stop-timeout=123.3',
            'd': '--stop-timeout=2.5',
            'e': '--stop-timeout=10.0',
            'f': '--stop-timeout=90.0',
            'g': '--stop-timeout=9120.0',
            'h': '--stop-timeout=20096.0',
            'i': '--stop-timeout=3661.001001',
        }

        for container, arg in result.items():
            cmd = []
            builder.container_run_args(cmd, container)
            self.assertIn(arg, cmd)

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_container_run_args_lists(self, runner):
        config = {
            'one': {
                'image': 'centos:7',
                'detach': False,
                'command': 'ls -l /foo',
                'remove': True,
                'tty': True,
                'interactive': True,
                'environment': ['FOO=BAR', 'BAR=BAZ'],
                'env_file': ['/tmp/foo.env', '/tmp/bar.env'],
                'ulimit': ['nofile=1024', 'nproc=1024'],
                'volumes': ['/foo:/foo:rw', '/bar:/bar:ro'],
                'volumes_from': ['two', 'three'],
                'group_add': ['docker', 'zuul'],
                'cap_add': ['SYS_ADMIN', 'SETUID'],
                'cap_drop': ['NET_RAW']
            }
        }
        builder = compose1.ComposeV1Builder('foo', config, runner)

        cmd = ['docker', 'run', '--name', 'one']
        builder.container_run_args(cmd, 'one')
        self.assertEqual(
            ['docker', 'run', '--name', 'one',
             '--env-file=/tmp/foo.env', '--env-file=/tmp/bar.env',
             '--env=FOO=BAR', '--env=BAR=BAZ',
             '--rm', '--interactive', '--tty',
             '--ulimit=nofile=1024', '--ulimit=nproc=1024',
             '--group-add=docker', '--group-add=zuul',
             '--volume=/foo:/foo:rw', '--volume=/bar:/bar:ro',
             '--volumes-from=two', '--volumes-from=three',
             '--cap-add=SYS_ADMIN', '--cap-add=SETUID', '--cap-drop=NET_RAW',
             'centos:7', 'ls', '-l', '/foo'],
            cmd
        )

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_container_run_args_lists_with_cpu_and_dict_env(self, runner):
        config = {
            'one': {
                'image': 'centos:7',
                'detach': False,
                'command': 'ls -l /foo',
                'remove': True,
                'tty': True,
                'interactive': True,
                'environment': {'BAR': 'BAZ', 'FOO': 'BAR', 'SINGLE': ''},
                'env_file': ['/tmp/foo.env', '/tmp/bar.env'],
                'ulimit': ['nofile=1024', 'nproc=1024'],
                'volumes': ['/foo:/foo:rw', '/bar:/bar:ro'],
                'volumes_from': ['two', 'three'],
                'group_add': ['docker', 'zuul'],
                'cap_add': ['SYS_ADMIN', 'SETUID'],
                'cap_drop': ['NET_RAW'],
                'cpuset_cpus': '0-2',
            }
        }
        builder = compose1.ComposeV1Builder('foo', config, runner)

        cmd = ['docker', 'run', '--name', 'one']
        builder.container_run_args(cmd, 'one')
        self.assertEqual(
            ['docker', 'run', '--name', 'one',
             '--env-file=/tmp/foo.env', '--env-file=/tmp/bar.env',
             '--env=BAR=BAZ', '--env=FOO=BAR', '--env=SINGLE',
             '--rm', '--interactive', '--tty',
             '--ulimit=nofile=1024', '--ulimit=nproc=1024',
             '--group-add=docker', '--group-add=zuul',
             '--volume=/foo:/foo:rw', '--volume=/bar:/bar:ro',
             '--volumes-from=two', '--volumes-from=three',
             '--cpuset-cpus=0-2',
             '--cap-add=SYS_ADMIN', '--cap-add=SETUID', '--cap-drop=NET_RAW',
             'centos:7', 'ls', '-l', '/foo'],
            cmd
        )

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_cont_exec_args(self, runner):
        r = runner.return_value
        r.discover_container_name.return_value = 'one-12345678'
        config = {
            'one': {
                'command': 'ls -l /foo',
                'privileged': True,
                'environment': {'FOO': 'BAR'},
                'user': 'bar'
            }
        }
        self.builder = compose1.ComposeV1Builder(
            'foo', config, runner.return_value)

        cmd = ['docker', 'exec']
        self.builder.cont_exec_args(cmd, 'one', 'one-12345678')
        self.assertEqual(
            ['docker', 'exec',
             '--privileged=true', '--user=bar',
             '--env=FOO=BAR',
             'one-12345678', '-l', '/foo'],
            cmd
        )

    def test_command_argument(self):
        b = compose1.ComposeV1Builder
        self.assertEqual([], b.command_argument(None))
        self.assertEqual([], b.command_argument(''))
        self.assertEqual([], b.command_argument([]))
        self.assertEqual(
            ['ls', '-l', '/foo-bar'],
            b.command_argument(['ls', '-l', '/foo-bar'])
        )
        self.assertEqual(
            ['ls', '-l', '/foo-bar'],
            b.command_argument('ls -l /foo-bar')
        )
        self.assertEqual(
            ['ls', '-l', '/foo bar'],
            b.command_argument(['ls', '-l', '/foo bar'])
        )
        # don't expect quoted spaces to do the right thing
        self.assertEqual(
            ['ls', '-l', '"/foo', 'bar"'],
            b.command_argument('ls -l "/foo bar"')
        )

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_delete_updated_no_change(self, runner):
        mock_inspect = mock.MagicMock()
        mock_inspect.side_effect = [
            {
                "Id": ("d038dccebdb0996ed36ab4ff06e7c424b3816d67664aa11e00642"
                       "be5e00cec55"),
                "Config": {
                    "Labels": {
                        "config_data": """{
                            \"start_order\": 0,
                            \"image": \"centos:7\"
                        }"""
                    },
                    "Image": "127.0.0.1:8787/centos:7"
                },
                "Image": "sha256:1"
            },
            "sha256:1"
        ]
        runner.inspect = mock_inspect
        mock_remove = mock.MagicMock()
        runner.remove_container = mock_remove

        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            }
        }

        self.builder = compose1.ComposeV1Builder(
            'one', config, runner.return_value)

        self.builder.runner = runner
        self.assertFalse(self.builder.delete_updated('one', [['one']]))
        calls = [
            mock.call('one'),
            mock.call('127.0.0.1:8787/centos:7',
                      '{{.Id}}',
                      o_type='image')
        ]
        mock_inspect.has_calls(calls)
        mock_remove.assert_not_called()

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_delete_updated_inspect_empty(self, runner):
        mock_inspect = mock.MagicMock()
        mock_inspect.return_value = None
        runner.inspect = mock_inspect
        mock_remove = mock.MagicMock()
        runner.remove_container = mock_remove

        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            }
        }

        self.builder = compose1.ComposeV1Builder(
            'one', config, runner.return_value)

        self.builder.runner = runner
        self.assertTrue(self.builder.delete_updated('one', [['one']]))
        mock_inspect.assert_called_once_with('one')
        mock_remove.assert_called_once_with('one')

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_delete_updated_no_config_data(self, runner):
        mock_inspect = mock.MagicMock()
        mock_inspect.side_effect = [
            {
                "Id": ("d038dccebdb0996ed36ab4ff06e7c424b3816d67664aa11e00642"
                       "be5e00cec55"),
                "Config": {
                    "Labels": {},
                    "Image": "127.0.0.1:8787/centos:7"
                },
                "Image": "sha256:1"
            },
            "sha256:1"
        ]
        runner.inspect = mock_inspect
        mock_remove = mock.MagicMock()
        runner.remove_container = mock_remove

        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            }
        }

        self.builder = compose1.ComposeV1Builder(
            'one', config, runner.return_value)

        self.builder.runner = runner
        self.assertTrue(self.builder.delete_updated('one', [['one']]))
        mock_inspect.assert_called_once_with('one')
        mock_remove.assert_called_once_with('one')

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_delete_updated_update_config(self, runner):
        mock_inspect = mock.MagicMock()
        mock_inspect.side_effect = [
            {
                "Id": ("d038dccebdb0996ed36ab4ff06e7c424b3816d67664aa11e00642"
                       "be5e00cec55"),
                "Config": {
                    "Labels": {
                        "config_data": """{
                            \"start_order\": 1,
                            \"image": \"centos:7\"
                        }"""
                    },
                    "Image": "127.0.0.1:8787/centos:7"
                },
                "Image": "sha256:1"
            },
            "sha256:1"
        ]

        runner.inspect = mock_inspect
        mock_remove = mock.MagicMock()
        runner.remove_container = mock_remove

        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            }
        }

        self.builder = compose1.ComposeV1Builder(
            'one', config, runner.return_value)

        self.builder.runner = runner
        self.assertTrue(self.builder.delete_updated('one', [['one']]))
        mock_inspect.assert_called_once_with('one')
        mock_remove.assert_called_once_with('one')

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_delete_updated_update_image(self, runner):
        mock_inspect = mock.MagicMock()
        mock_inspect.side_effect = [
            {
                "Id": ("d038dccebdb0996ed36ab4ff06e7c424b3816d67664aa11e00642"
                       "be5e00cec55"),
                "Config": {
                    "Labels": {
                        "config_data": """{
                            \"start_order\": 0,
                            \"image": \"centos:7\"
                        }"""
                    },
                    "Image": "127.0.0.1:8787/centos:7"
                },
                "Image": "sha256:1"
            },
            "sha256:2"
        ]

        runner.inspect = mock_inspect
        mock_remove = mock.MagicMock()
        runner.remove_container = mock_remove

        config = {
            'one': {
                'start_order': 0,
                'image': 'centos:7',
            }
        }

        self.builder = compose1.ComposeV1Builder(
            'one', config, runner.return_value)

        self.builder.runner = runner
        self.assertTrue(self.builder.delete_updated('one', [['one']]))
        calls = [
            mock.call('one'),
            mock.call('127.0.0.1:8787/centos:7',
                      '{{.Id}}',
                      o_type='image')
        ]
        mock_inspect.has_calls(calls)
        mock_remove.assert_called_once_with('one')


class TestVolumeChecks(base.TestCase):
    @mock.patch('paunch.runner.PodmanRunner', autospec=True)
    def test_validate_volumes(self, runner):
        runner.validate_volume_source.return_value = True
        builder = basebuilder.BaseBuilder('test', {}, runner, {})
        volumes = ['', '/var:/var', 'test:/bar']
        self.assertTrue(builder.validate_volumes(volumes))

    def test_validate_volumes_empty(self):
        builder = basebuilder.BaseBuilder('test', {}, runner, {})
        volumes = []
        self.assertTrue(builder.validate_volumes(volumes))

    @mock.patch('paunch.runner.PodmanRunner', autospec=True)
    def test_validate_volumes_fail(self, runner):
        runner.validate_volume_source.return_value = False
        builder = basebuilder.BaseBuilder('test', {}, runner, {})
        volumes = ['/var:/var']
        self.assertFalse(builder.validate_volumes(volumes))



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil

from paunch import constants
from paunch.utils import common
from paunch.utils import systemctl

DROP_IN_MARKER_FILE = '/etc/sysconfig/podman_drop_in'


def service_create(container, cconfig, sysdir=constants.SYSTEMD_DIR, log=None):
    """Create a service in systemd

    :param container: container name
    :type container: String

    :param cconfig: container configuration
    :type cconfig: Dictionary

    :param sysdir: systemd unit files directory
    :type sysdir: String

    :param log: optional pre-defined logger for messages
    :type log: logging.RootLogger
    """
    log = log or common.configure_logging(__name__)
    # We prefix the SystemD service so we can identify them better:
    # e.g. systemctl list-unit-files | grep tripleo_
    # It'll help to not conflict when rpms are installed on the host and
    # have the same service name as their container name.
    # For example haproxy rpm and haproxy container would have the same
    # service name so the prefix will help to not having this conflict
    # when removing the rpms during a cleanup by the operator.
    service = 'tripleo_' + container

    wants = " ".join(systemctl.format_name(str(x)) for x in
                     cconfig.get('depends_on', []))

    restart = cconfig.get('restart', 'always')
    stop_grace_period = cconfig.get('stop_grace_period', '10')

    # Please refer to systemd.exec documentation for those entries
    # https://www.freedesktop.org/software/systemd/man/systemd.exec.html
    sys_exec = cconfig.get('systemd_exec_flags', {})

    # SystemD doesn't have the equivalent of docker unless-stopped.
    # Let's force 'always' so containers aren't restarted when stopped by
    # systemd, but restarted when in failure. Also this code is only for
    # podman now, so nothing changed for Docker deployments.
    if restart == 'unless-stopped':
        restart = 'always'

    # If the service depends on other services, it must be stopped
    # in a specific order. The host can be configured to prevent
    # systemd from stopping the associated systemd scopes too early,
    # so make sure to generate the start command accordingly.
    if (len(cconfig.get('depends_on', [])) > 0 and
            os.path.exists(DROP_IN_MARKER_FILE)):
        start_cmd = '/usr/libexec/paunch-start-podman-container %s' % container
    else:
        start_cmd = '/usr/bin/podman start %s' % container

    sysd_unit_f = sysdir + systemctl.format_name(service)
    log.debug('Creating systemd unit file: %s' % sysd_unit_f)
    s_config = {
        'name': container,
        'start_cmd': start_cmd,
        'wants': wants,
        'restart': restart,
        'stop_grace_period': stop_grace_period,
        'sys_exec': '\n'.join(['%s=%s' % (x, y) for x, y in sys_exec.items()]),
    }
    # Ensure we don't have some trailing .requires directory and content for
    # this service
    if os.path.exists(sysd_unit_f + '.requires'):
        shutil.rmtree(sysd_unit_f + '.requires')

    with open(sysd_unit_f, 'w') as unit_file:
        os.chmod(unit_file.name, 0o644)
        unit_file.write("""[Unit]
Description=%(name)s container
After=paunch-container-shutdown.service
Wants=%(wants)s
[Service]
Restart=%(restart)s
ExecStart=%(start_cmd)s
ExecReload=/usr/bin/podman kill --signal HUP %(name)s
ExecStop=/usr/bin/podman stop -t %(stop_grace_period)s %(name)s
ExecStopPost=/usr/bin/podman stop -t %(stop_grace_period)s %(name)s
SuccessExitStatus=137 142 143
KillMode=none
Type=forking
PIDFile=/var/run/%(name)s.pid
%(sys_exec)s
[Install]
WantedBy=multi-user.target""" % s_config)
    try:
        systemctl.daemon_reload()
        systemctl.enable(service, now=True)
    except systemctl.SystemctlException:
        log.exception("systemctl failed")
        raise


def service_delete(container, sysdir=constants.SYSTEMD_DIR, log=None):
    """Delete a service in systemd

    :param container: container name
    :type container: String

    :param sysdir: systemd unit files directory
    :type sysdir: string

    :param log: optional pre-defined logger for messages
    :type log: logging.RootLogger
    """
    log = log or common.configure_logging(__name__)
    # prefix is explained in the service_create().
    service = 'tripleo_' + container

    sysd_unit_f = systemctl.format_name(service)
    sysd_health_f = systemctl.format_name(service + '_healthcheck')
    sysd_timer_f = service + '_healthcheck.timer'
    sysd_health_req_d = sysd_unit_f + '.requires'

    for sysd_f in sysd_unit_f, sysd_health_f, sysd_timer_f:
        if os.path.isfile(sysdir + sysd_f):
            log.debug('Stopping and disabling systemd service for %s' %
                      service)
            try:
                systemctl.stop(sysd_f)
                systemctl.disable(sysd_f)
            except systemctl.SystemctlException:
                log.exception("systemctl failed")
                raise
            log.debug('Removing systemd unit file %s' % sysd_f)
            os.remove(sysdir + sysd_f)
        else:
            log.info('No systemd unit file was found for %s' % sysd_f)

    # Now that the service is removed, we can remove its ".requires"
    if os.path.exists(os.path.join(sysdir, sysd_health_req_d)):
            log.info('Removing healthcheck require for %s' % service)
            shutil.rmtree(os.path.join(sysdir, sysd_health_req_d))


def healthcheck_create(container, sysdir='/etc/systemd/system/',
                       log=None, test='/openstack/healthcheck'):
    """Create a healthcheck for a service in systemd

    :param container: container name
    :type container: String

    :param sysdir: systemd unit files directory
    :type sysdir: String

    :param log: optional pre-defined logger for messages
    :type log: logging.RootLogger

    :param test: optional test full command
    :type test: String
    """

    log = log or common.configure_logging(__name__)

    service = 'tripleo_' + container
    healthcheck = systemctl.format_name(service + '_healthcheck')
    sysd_unit_f = sysdir + healthcheck
    log.debug('Creating systemd unit file: %s' % sysd_unit_f)
    s_config = {
        'name': container,
        'service': service,
        'restart': 'restart',
        'test': test,
    }
    with open(sysd_unit_f, 'w') as unit_file:
        os.chmod(unit_file.name, 0o644)
        unit_file.write("""[Unit]
Description=%(name)s healthcheck
After=paunch-container-shutdown.service %(service)s.service
Requisite=%(service)s.service
[Service]
Type=oneshot
ExecStart=/usr/bin/podman exec --user root %(name)s %(test)s
SyslogIdentifier=healthcheck_%(name)s
[Install]
WantedBy=multi-user.target
""" % s_config)


def healthcheck_timer_create(container, cconfig, sysdir='/etc/systemd/system/',
                             log=None):
    """Create a systemd timer for a healthcheck

    :param container: container name
    :type container: String

    :param cconfig: container configuration
    :type cconfig: Dictionary

    :param sysdir: systemd unit files directory
    :type sysdir: string

    :param log: optional pre-defined logger for messages
    :type log: logging.RootLogger
    """

    log = log or common.configure_logging(__name__)

    service = 'tripleo_' + container
    healthcheck_timer = service + '_healthcheck.timer'
    sysd_timer_f = sysdir + healthcheck_timer
    log.debug('Creating systemd timer file: %s' % sysd_timer_f)
    interval = cconfig.get('check_interval', 60)
    s_config = {
        'name': container,
        'service': service,
        'interval': interval,
        'randomize': int(interval) * 3 / 4
    }
    with open(sysd_timer_f, 'w') as timer_file:
        os.chmod(timer_file.name, 0o644)
        timer_file.write("""[Unit]
Description=%(name)s container healthcheck
PartOf=%(service)s.service
[Timer]
OnActiveSec=120
OnUnitActiveSec=%(interval)s
RandomizedDelaySec=%(randomize)s
[Install]
WantedBy=timers.target""" % s_config)
    try:
        systemctl.enable(healthcheck_timer, now=True)
        systemctl.add_requires(systemctl.format_name(service),
                               healthcheck_timer)
        systemctl.daemon_reload()
    except systemctl.SystemctlException:
        log.exception("systemctl failed")
        raise



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

import paunch


"""Utility to launch and manage containers using
YAML based configuration data"""


class PaunchApp(App):

    def __init__(self):
        super(PaunchApp, self).__init__(
            description=__doc__,
            version=paunch.__version__,
            command_manager=CommandManager('paunch'),
            deferred_help=True,
            )


def main(argv=sys.argv[1:]):
    myapp = PaunchApp()
    return myapp.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import collections
import jmespath
import json
import os
import random
import string
import subprocess
import time

from paunch.builder import podman
from paunch.utils import common
from paunch.utils import systemctl
from paunch.utils import systemd


class BaseRunner(object):
    def __init__(self, managed_by, cont_cmd, log=None, cont_log_path=None,
                 healthcheck_disabled=False):
        self.managed_by = managed_by
        self.cont_cmd = cont_cmd
        # Leverage pre-configured logger
        self.log = log or common.configure_logging(__name__)
        self.cont_log_path = cont_log_path
        self.healthcheck_disabled = healthcheck_disabled
        if self.cont_cmd == 'docker':
            self.log.warning('docker runtime is deprecated in Stein '
                             'and will be removed in Train.')

    @staticmethod
    def execute(cmd, log=None, quiet=False, warn_only=False):
        if not log:
            log = common.configure_logging(__name__)
        if not quiet:
            log.debug('$ %s' % ' '.join(cmd))
        subproc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        cmd_stdout, cmd_stderr = subproc.communicate()
        if subproc.returncode != 0:
            if warn_only:
                log.warning('Error executing %s: '
                            'returned %s' % (cmd, subproc.returncode))
            else:
                log.error('Error executing %s: '
                          'returned %s' % (cmd, subproc.returncode))
        if not quiet:
            log.debug(cmd_stdout)
            log.debug(cmd_stderr)
        return (cmd_stdout.decode('utf-8'),
                cmd_stderr.decode('utf-8'),
                subproc.returncode)

    @staticmethod
    def execute_interactive(cmd, log=None):
        if not log:
            log = common.configure_logging(__name__)
        log.debug('$ %s' % ' '.join(cmd))
        return subprocess.call(cmd)

    def current_config_ids(self):
        # List all config_id labels for managed containers
        # FIXME(bogdando): remove once we have it fixed:
        # https://github.com/containers/libpod/issues/1729
        if self.cont_cmd == 'docker':
            fmt = '{{.Label "config_id"}}'
        else:
            fmt = '{{.Labels.config_id}}'
        cmd = [
            self.cont_cmd, 'ps', '-a',
            '--filter', 'label=managed_by=%s' % self.managed_by,
            '--format', fmt
        ]
        cmd_stdout, cmd_stderr, returncode = self.execute(
            cmd, log=self.log, quiet=False, warn_only=True)
        results = cmd_stdout.split()
        if returncode != 0 or not results or results == ['']:
            # NOTE(bogdando): also look by the historically used to
            # be always specified defaults, we must also identify such configs
            cmd = [
                self.cont_cmd, 'ps', '-a',
                '--filter', 'label=managed_by=paunch',
                '--format', fmt
            ]
            cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
            if returncode != 0:
                return set()
            results += cmd_stdout.split()
        return set(results)

    def containers_in_config(self, conf_id):
        cmd = [
            self.cont_cmd, 'ps', '-q', '-a',
            '--filter', 'label=managed_by=%s' % self.managed_by,
            '--filter', 'label=config_id=%s' % conf_id
        ]
        cmd_stdout, cmd_stderr, returncode = self.execute(
            cmd, log=self.log, quiet=False, warn_only=True)
        results = cmd_stdout.split()
        if returncode != 0 or not results or results == ['']:
            # NOTE(bogdando): also look by the historically used to
            # be always specified defaults, we must also identify such configs
            cmd = [
                self.cont_cmd, 'ps', '-q', '-a',
                '--filter', 'label=managed_by=paunch',
                '--filter', 'label=config_id=%s' % conf_id
            ]
            cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
            if returncode != 0:
                return []
            results += cmd_stdout.split()

        return [c for c in results]

    def inspect(self, name, output_format=None, o_type='container',
                quiet=False):
        # In podman, if we're being asked to inspect a container image, we
        # want to verify that the image exists before inspecting it.
        # Context: https://github.com/containers/libpod/issues/1845
        if o_type == 'image':
            if not self.image_exist(name):
                return
        cmd = [self.cont_cmd, 'inspect', '--type', o_type]
        if output_format:
            cmd.append('--format')
            cmd.append(output_format)
        cmd.append(name)
        (cmd_stdout, cmd_stderr, returncode) = self.execute(
            cmd, self.log, quiet, True)
        if returncode != 0:
            return
        try:
            if output_format:
                return cmd_stdout
            else:
                return json.loads(cmd_stdout)[0]
        except Exception as e:
            self.log.error('Problem parsing %s inspect: %s' %
                           (self.cont_cmd, e))

    def unique_container_name(self, container):
        container_name = container
        if self.cont_cmd == 'docker':
            while self.inspect(container_name, output_format='exists',
                               quiet=True):
                suffix = ''.join(random.choice(
                    string.ascii_lowercase + string.digits) for i in range(8))
                container_name = '%s-%s' % (container, suffix)
                break
        else:
            while self.container_exist(container_name, quiet=True):
                suffix = ''.join(random.choice(
                    string.ascii_lowercase + string.digits) for i in range(8))
                container_name = '%s-%s' % (container, suffix)
                break
        return container_name

    def discover_container_name(self, container, cid):
        cmd = [
            self.cont_cmd,
            'ps',
            '-a',
            '--filter',
            'label=container_name=%s' % container,
            '--filter',
            'label=config_id=%s' % cid,
            '--format',
            '{{.Names}}'
        ]
        (cmd_stdout, cmd_stderr, returncode) = self.execute(
            cmd, log=self.log, quiet=False, warn_only=True)
        if returncode == 0:
            names = cmd_stdout.split()
            if names:
                return names[0]
        self.log.warning('Did not find container with "%s" - retrying without '
                         'config_id' % cmd)

        cmd = [
            self.cont_cmd,
            'ps',
            '-a',
            '--filter',
            'label=container_name=%s' % container,
            '--format',
            '{{.Names}}'
        ]
        (cmd_stdout, cmd_stderr, returncode) = self.execute(cmd, self.log)
        if returncode == 0:
            names = cmd_stdout.split()
            if names:
                return names[0]

        self.log.warning('Did not find container with "%s"' % cmd)

    def delete_missing_configs(self, config_ids):
        if not config_ids:
            config_ids = []

        for conf_id in self.current_config_ids():
            if conf_id not in config_ids:
                self.log.debug('%s no longer exists, deleting containers' %
                               conf_id)
                self.remove_containers(conf_id)

    def discover_container_config(self, configs, container, name):
        '''Find the paunch and runtime configs of a container by name.'''
        for conf_id in self.current_config_ids():
            jquerry = ("[] | [?(Name=='%s' && "
                       "Config.Labels.container_name=='%s' && "
                       "Config.Labels.config_id=='%s')]" %
                       (container, name, conf_id))
            runtime_conf = None
            try:
                runtime_conf = jmespath.search(jquerry,
                                               configs[conf_id])[0]
                result = (conf_id, runtime_conf)
            except Exception:
                self.log.error("Failed searching container %s "
                               "for config %s" % (container, conf_id))
                result = (None, None)
            if runtime_conf:
                self.log.debug("Found container %s "
                               "for config %s" % (container, conf_id))
                break
        return result

    def list_configs(self):
        configs = collections.defaultdict(list)
        for conf_id in self.current_config_ids():
            for container in self.containers_in_config(conf_id):
                configs[conf_id].append(self.inspect(container))
        return configs

    def container_names(self, conf_id=None):
        # list every container name, and its container_name label
        # FIXME(bogdando): remove once we have it fixed:
        # https://github.com/containers/libpod/issues/1729
        if self.cont_cmd == 'docker':
            fmt = '{{.Label "container_name"}}'
        else:
            fmt = '{{.Labels.container_name}}'
        cmd = [
            self.cont_cmd, 'ps', '-a',
            '--filter', 'label=managed_by=%s' % self.managed_by
        ]
        if conf_id:
            cmd.extend((
                '--filter', 'label=config_id=%s' % conf_id
            ))
        cmd.extend((
            '--format', '{{.Names}} %s' % fmt
        ))
        cmd_stdout, cmd_stderr, returncode = self.execute(
            cmd, log=self.log, quiet=False, warn_only=True)
        results = cmd_stdout.split("\n")
        if returncode != 0 or not results or results == ['']:
            # NOTE(bogdando): also look by the historically used to
            # be always specified defaults, we must also identify such configs
            cmd = [
                self.cont_cmd, 'ps', '-a',
                '--filter', 'label=managed_by=paunch'
            ]
            if conf_id:
                cmd.extend((
                    '--filter', 'label=config_id=%s' % conf_id
                ))
            cmd.extend((
                '--format', '{{.Names}} %s' % fmt
            ))
            cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
            if returncode != 0:
                return []
            results += cmd_stdout.split("\n")
        result = []
        for line in results:
            if line:
                result.append(line.split())
        return result

    def remove_containers(self, conf_id):
        for container in self.containers_in_config(conf_id):
            self.remove_container(container)

    def remove_container(self, container):
        self.execute([self.cont_cmd, 'stop', container], self.log)
        cmd = [self.cont_cmd, 'rm', container]
        cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
        if returncode != 0:
            self.log.error('Error removing container '
                           'gracefully: %s' % container)
            self.log.error(cmd_stderr)
            cmd = [self.cont_cmd, 'rm', '-f', container]
            cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
            if returncode != 0:
                self.log.error('Error removing container: %s' % container)
                self.log.error(cmd_stderr)
                raise Exception('Unable to remove container: %s' % container)

    def stop_container(self, container, cont_cmd=None, quiet=False):
        cont_cmd = cont_cmd or self.cont_cmd
        cmd = [cont_cmd, 'stop', container]
        cmd_stdout, cmd_stderr, returncode = self.execute(cmd, quiet=quiet)
        if returncode != 0 and not quiet:
            self.log.error('Error stopping container: %s' % container)
            self.log.error(cmd_stderr)
            raise Exception('Unable to stop container: %s' % container)

    def rename_containers(self):
        current_containers = []
        need_renaming = {}
        renamed = False
        for entry in self.container_names():
            current_containers.append(entry[0])

            # ignore if container_name label not set
            if len(entry) < 2:
                continue

            # ignore if desired name is already actual name
            if entry[0] == entry[-1]:
                continue

            need_renaming[entry[0]] = entry[-1]

        for current, desired in sorted(need_renaming.items()):
            if desired in current_containers:
                self.log.info('Cannot rename "%s" since "%s" still exists' % (
                    current, desired))
            else:
                self.log.info('Renaming "%s" to "%s"' % (current, desired))
                self.rename_container(current, desired)
                renamed = True
                current_containers.append(desired)
        return renamed

    def validate_volume_source(self, volume):
        """Validate that the provided volume

        This checks that the provided volume either exists on the filesystem
        or is a container volume.

        :param: volume: string containing either a filesystme path or container
                        volume name
        """
        if os.path.exists(volume):
            return True

        if os.path.sep in volume:
            # if we get here and have a path seperator, let's skip the
            # container lookup because container volumes won't have / in them.
            self.log.debug('Path seperator found in volume (%s), but did not '
                           'exist on the file system' % volume)
            return False

        self.log.debug('Running volume lookup for "%s"' % volume)
        filter_opt = '--filter=name={}'.format(volume)
        cmd = [self.cont_cmd, 'volume', 'ls', '-q', filter_opt]
        cmd_stdout, cmd_stderr, returncode = self.execute(cmd)
        if returncode != 0:
            self.log.error('Error during volume verification')
            self.log.error(cmd_stderr)
            return False
        return (volume in set(cmd_stdout.split()))


class DockerRunner(BaseRunner):

    def __init__(self, managed_by, cont_cmd=None, log=None):
        cont_cmd = cont_cmd or 'docker'
        super(DockerRunner, self).__init__(managed_by, cont_cmd, log)

    def rename_container(self, container, name):
        cmd = [self.cont_cmd, 'rename', container, name]
        cmd_stdout, cmd_stderr, returncode = self.execute(cmd, self.log)
        if returncode != 0:
            self.log.error('Error renaming container: %s' % container)
            self.log.error(cmd_stderr)

    def image_exist(self, name, quiet=False):
        self.log.warning("image_exist isn't supported "
                         "by %s" % self.cont_cmd)
        return True

    def container_exist(self, name, quiet=False):
        self.log.warning("container_exist isn't supported "
                         "by %s" % self.cont_cmd)
        return True

    def container_running(self, container):
        self.log.warning("container_running isn't supported "
                         "by %s" % self.cont_cmd)
        return True


class PodmanRunner(BaseRunner):

    def __init__(self, managed_by, cont_cmd=None, log=None,
                 cont_log_path=None, healthcheck_disabled=False):
        cont_cmd = cont_cmd or 'podman'
        super(PodmanRunner, self).__init__(managed_by, cont_cmd, log,
                                           cont_log_path, healthcheck_disabled)

    def rename_container(self, container, name):
        # TODO(emilien) podman doesn't support rename, we'll handle it
        # in paunch itself for now
        configs = self.list_configs()
        config_id, config = self.discover_container_config(
            configs, container, name)
        # Get config_data dict by the discovered conf ID,
        # paunch needs it for maintaining idempotency within a conf ID
        filter_names = ("[] | [?(Name!='%s' && "
                        "Config.Labels.config_id=='%s')]"
                        ".Name" % (container, config_id))
        filter_cdata = ("[] | [?(Name!='%s' && "
                        "Config.Labels.config_id=='%s')]"
                        ".Config.Labels.config_data" % (container, config_id))
        names = None
        cdata = None
        try:
            names = jmespath.search(filter_names, configs[config_id])
            cdata = jmespath.search(filter_cdata, configs[config_id])
        except jmespath.exceptions.LexerError:
            self.log.error("Failed to rename a container %s into %s: "
                           "used a bad search pattern" % (container, name))
            return

        if not names or not cdata:
            self.log.error("Failed to rename a container %s into %s: "
                           "no config_data was found" % (container, name))
            return

        # Rename the wanted container in the config_data fetched from the
        # discovered config
        config_data = dict(zip(names, map(json.loads, cdata)))
        config_data[name] = json.loads(
            config.get('Config').get('Labels').get('config_data'))

        # Re-apply a container under its amended name using the fetched configs
        self.log.debug("Renaming a container known as %s into %s, "
                       "via re-applying its original config" %
                       (container, name))
        # destination container
        self.stop_container(name)
        self.remove_container(name)
        self.stop_container(container)
        self.remove_container(container)
        builder = podman.PodmanBuilder(
            config_id=config_id,
            config=config_data,
            runner=self,
            labels=None,
            log=self.log,
            cont_log_path=self.cont_log_path,
            healthcheck_disabled=self.healthcheck_disabled
        )
        builder.apply()

    def stop_container(self, container, cont_cmd=None, quiet=False):
        if not self.container_running(container):
            self.log.debug('%s not running, skipping stop' % container)
            return
        self.log.debug("Stopping container: %s" % container)
        return super(PodmanRunner, self).stop_container(container,
                                                        cont_cmd,
                                                        quiet)

    def remove_container(self, container):
        if not self.container_exist(container):
            self.log.debug('%s does not exist, skipping remove' % container)
            return
        self.log.debug("Removing container: %s" % container)
        systemd.service_delete(container=container, log=self.log)
        return super(PodmanRunner, self).remove_container(container)

    def image_exist(self, name, quiet=False):
        cmd = ['podman', 'image', 'exists', name]
        (_, _, returncode) = self.execute(cmd, self.log, quiet, True)
        return returncode == 0

    def container_exist(self, name, quiet=False):
        cmd = ['podman', 'container', 'exists', name]
        (_, _, returncode) = self.execute(cmd, self.log, quiet, True)
        return returncode == 0

    def container_running(self, container):
        if not self.container_exist(container):
            self.log.debug('%s is not running because it does not exist' %
                           container)
            return False
        service_name = 'tripleo_' + container + '.service'
        try:
            systemctl.is_active(service_name)
            self.log.debug('Unit %s is running' % service_name)
            return True
        except systemctl.SystemctlException:
            chk_cmd = [
                self.cont_cmd,
                'ps',
                '--filter',
                'label=container_name=%s' % container,
                '--format',
                '{{.Names}}'
            ]
            cmd_stdout = ''
            returncode = -1
            count = 1
            while (not cmd_stdout or returncode != 0) and count <= 5:
                self.log.warning('Attempt %i to check if %s is '
                                 'running' % (count, container))
                # at the first retry, we will force a sync with the OCI runtime
                if self.cont_cmd == 'podman' and count == 2:
                    chk_cmd.append('--sync')
                (cmd_stdout, cmd_stderr, returncode) = self.execute(
                    chk_cmd, log=self.log, quiet=False, warn_only=True)

                if returncode != 0:
                    self.log.warning('Attempt %i Error when running '
                                     '%s:' % (count, chk_cmd))
                    self.log.warning(cmd_stderr)
                else:
                    if not cmd_stdout:
                        self.log.warning('Attempt %i Container %s '
                                         'is not running' % (count, container))

                count += 1
                time.sleep(0.2)
            # return True if ps ran successfuly and returned a container name.
            return (cmd_stdout and returncode == 0)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import collections

from cliff import command
from cliff import lister
import json

import paunch

from paunch import utils


class Apply(command.Command):

    log = None

    def get_parser(self, prog_name):
        parser = super(Apply, self).get_parser(prog_name)
        parser.add_argument(
            '--file',
            metavar='<file>',
            required=True,
            help=('YAML or JSON file or directory containing configuration '
                  'data'),
        )
        parser.add_argument(
            '--label',
            metavar='<label=value>',
            dest='labels',
            default=[],
            action='append',
            help=('Extra labels to apply to containers in this config, in the '
                  'form --label label=value --label label2=value2.'),
        )
        parser.add_argument(
            '--managed-by',
            metavar='<name>',
            dest='managed_by',
            default='paunch',
            help=('Override the name of the tool managing the containers'),
        )
        parser.add_argument(
            '--config-id',
            metavar='<name>',
            dest='config_id',
            required=True,
            help=('ID to assign to containers'),
        )
        parser.add_argument(
            '--default-runtime',
            dest='default_runtime',
            default='podman',
            choices=['docker', 'podman'],
            help=('Default runtime for containers. Can be docker or podman.'),
        )
        parser.add_argument(
            '--container-log-path',
            dest='cont_log_path',
            default=None,
            help=('Absolute directory path for container log. Works only for '
                  'podman container engine')
        )
        parser.add_argument(
            '--healthcheck-disabled',
            dest='healthcheck_disabled',
            action='store_true',
            default=False,
            help=('Whether or not we disable the containers healthchecks')
        )
        parser.add_argument(
            '--no-cleanup',
            dest='cleanup',
            action='store_false',
            default=True,
            help=('Whether or not we delete containers missing in the config')
        )
        return parser

    def take_action(self, parsed_args):
        (self.log, log_file, log_level) = \
            utils.common.configure_logging_from_args(__name__, self.app_args)
        labels = collections.OrderedDict()
        for l in parsed_args.labels:
            k, v = l.split(('='), 1)
            labels[k] = v

        config_path = parsed_args.file
        config = utils.common.load_config(config_path)

        stdout, stderr, rc = paunch.apply(
            parsed_args.config_id,
            config,
            managed_by=parsed_args.managed_by,
            labels=labels,
            cont_cmd=parsed_args.default_runtime,
            log_level=log_level,
            log_file=log_file,
            cont_log_path=parsed_args.cont_log_path,
            healthcheck_disabled=parsed_args.healthcheck_disabled,
            cleanup=parsed_args.cleanup
        )

        return rc


class Cleanup(command.Command):

    log = None

    def get_parser(self, prog_name):
        parser = super(Cleanup, self).get_parser(prog_name)
        parser.add_argument(
            'config_id',
            metavar='<config_id>',
            nargs='*',
            help=('Identifiers for the configs which still apply, all others '
                  'will be deleted.'),
        )
        parser.add_argument(
            '--managed-by',
            metavar='<name>',
            dest='managed_by',
            default='paunch',
            help=('Override the name of the tool managing the containers'),
        )
        parser.add_argument(
            '--default-runtime',
            dest='default_runtime',
            default='podman',
            choices=['docker', 'podman'],
            help=('Default runtime for containers. Can be docker or podman.'),
        )
        return parser

    def take_action(self, parsed_args):
        (self.log, log_file, log_level) = \
            utils.common.configure_logging_from_args(__name__, self.app_args)
        paunch.cleanup(
            parsed_args.config_id,
            managed_by=parsed_args.managed_by,
            cont_cmd=parsed_args.default_runtime,
            log_level=log_level,
            log_file=log_file
        )


class Delete(command.Command):

    log = None

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)
        parser.add_argument(
            'config_id',
            nargs='*',
            metavar='<config_id>',
            help=('Identifier for the config to delete the containers for'),
        )
        parser.add_argument(
            '--managed-by',
            metavar='<name>',
            dest='managed_by',
            default='paunch',
            help=('Override the name of the tool managing the containers'),
        )
        parser.add_argument(
            '--default-runtime',
            dest='default_runtime',
            default='podman',
            choices=['docker', 'podman'],
            help=('Default runtime for containers. Can be docker or podman.'),
        )
        return parser

    def take_action(self, parsed_args):
        (self.log, log_file, log_level) = \
            utils.common.configure_logging_from_args(__name__, self.app_args)
        paunch.delete(
            parsed_args.config_id,
            parsed_args.managed_by,
            cont_cmd=parsed_args.default_runtime,
            log_level=log_level,
            log_file=log_file
        )


class Debug(command.Command):

    log = None

    def get_parser(self, prog_name):
        parser = super(Debug, self).get_parser(prog_name)
        parser.add_argument(
            '--file',
            metavar='<file>',
            required=True,
            help=('YAML or JSON file or directory containing configuration '
                  'data'),
        )
        parser.add_argument(
            '--label',
            metavar='<label=value>',
            dest='labels',
            default=[],
            action='append',
            help=('Extra labels to apply to containers in this config, in the '
                  'form --label label=value --label label2=value2.'),
        )
        parser.add_argument(
            '--managed-by',
            metavar='<name>',
            dest='managed_by',
            default='paunch',
            help=('Override the name of the tool managing the containers')
        )
        parser.add_argument(
            '--action',
            metavar='<name>',
            dest='action',
            default='print-cmd',
            help=('Action can be one of: "dump-json", "dump-yaml", '
                  '"print-cmd", or "run"')
        )
        parser.add_argument(
            '--container',
            metavar='<name>',
            dest='container_name',
            required=True,
            help=('Name of the container you wish to manipulate')
        )
        parser.add_argument(
            '--interactive',
            dest='interactive',
            action='store_true',
            default=False,
            help=('Run container in interactive mode - modifies config and '
                  'execution of container')
        )
        parser.add_argument(
            '--shell',
            dest='shell',
            action='store_true',
            default=False,
            help=('Similar to interactive but drops you into a shell')
        )
        parser.add_argument(
            '--user',
            metavar='<name>',
            dest='user',
            default='',
            help=('Start container as the specified user')
        )
        parser.add_argument(
            '--overrides',
            metavar='<name>',
            dest='overrides',
            default='',
            help=('JSON configuration information used to override default '
                  'config values')
        )
        parser.add_argument(
            '--config-id',
            metavar='<name>',
            dest='config_id',
            required=False,
            default='debug',
            help=('ID to assign to containers')
        )
        parser.add_argument(
            '--default-runtime',
            dest='default_runtime',
            default='podman',
            choices=['docker', 'podman'],
            help=('Default runtime for containers. Can be docker or podman.'),
        )
        return parser

    def take_action(self, parsed_args):
        (self.log, log_file, log_level) = \
            utils.common.configure_logging_from_args(__name__, self.app_args)
        labels = collections.OrderedDict()
        for l in parsed_args.labels:
            k, v = l.split(('='), 1)
            labels[k] = v

        container_name = parsed_args.container_name
        config_path = parsed_args.file
        config = utils.common.load_config(config_path, container_name)

        cconfig = {}
        cconfig[container_name] = config[container_name]

        if parsed_args.interactive or parsed_args.shell:
            iconfig = {
                "interactive": True,
                "tty": True,
                "restart": "no",
                "detach": False,
                "remove": True
                }
            cconfig[container_name].update(iconfig)
        if parsed_args.shell:
            sconfig = {"command": "/bin/bash"}
            cconfig[container_name].update(sconfig)
        if parsed_args.user:
            rconfig = {"user": parsed_args.user}
            cconfig[container_name].update(rconfig)

        conf_overrides = []
        if parsed_args.overrides:
            conf_overrides = json.loads(parsed_args.overrides)

        cconfig[container_name].update(conf_overrides)

        paunch.debug(
            parsed_args.config_id,
            container_name,
            parsed_args.action,
            cconfig,
            parsed_args.managed_by,
            labels=labels,
            cont_cmd=parsed_args.default_runtime,
            log_level=log_level,
            log_file=log_file
        )


class List(lister.Lister):

    log = None

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)
        parser.add_argument(
            '--managed-by',
            metavar='<name>',
            dest='managed_by',
            default='paunch',
            help=('Override the name of the tool managing the containers'),
        )
        parser.add_argument(
            '--default-runtime',
            dest='default_runtime',
            default='podman',
            choices=['docker', 'podman'],
            help=('Default runtime for containers. Can be docker or podman.'),
        )
        return parser

    def take_action(self, parsed_args):
        (self.log, log_file, log_level) = \
            utils.common.configure_logging_from_args(__name__, self.app_args)
        configs = paunch.list(
            parsed_args.managed_by,
            cont_cmd=parsed_args.default_runtime,
            log_level=log_level,
            log_file=log_file
        )
        columns = [
            'config',
            'container',
            'image',
            'command',
            'status',
        ]
        data = []
        for k, v in configs.items():
            for i in v:
                # docker has a leading slash in the name, strip it
                if parsed_args.default_runtime == 'docker':
                    name = i.get('Name', '/')[1:]
                else:
                    name = i.get('Name', '')
                cmd = ' '.join(i.get('Config', {}).get('Cmd', []))
                image = i.get('Config', {}).get('Image')
                status = i.get('State', {}).get('Status')
                data.append((k, name, image, cmd, status))
        return columns, data



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''Stable library interface to managing containers with paunch.'''

import json
import pbr.version
import yaml

from paunch.builder import compose1
from paunch.builder import podman
from paunch import runner
from paunch.utils import common

__version__ = pbr.version.VersionInfo('paunch').version_string()


def apply(config_id, config, managed_by, labels=None, cont_cmd='podman',
          default_runtime=None, log_level=None, log_file=None,
          cont_log_path=None, healthcheck_disabled=False, cleanup=True):
    """Execute supplied container configuration.

    :param str config_id: Unique config ID, should not be re-used until any
                          running containers with that config ID have been
                          deleted.
    :param dict config: Configuration data describing container actions to
                        apply.
    :param str managed_by: Name of the tool managing the containers. Only
                           containers labelled with this will be modified.
    :param dict labels: Optional keys/values of labels to apply to containers
                        created with this invocation.
    :param str cont_cmd: Optional override to the container command to run.
    :param str default_runtime: (deprecated) does nothing.
    :param int log_level: optional log level for loggers
    :param str log_file: optional log file for messages
    :param str cont_log_path: optional log path for containers. Works only for
                              podman engine. Must be an absolute path.
    :param bool healthcheck_disabled: optional boolean to disable container
                                      healthcheck.
    :param bool cleanup: optional boolean to delete containers missing in the
                         config.

    :returns (list, list, int) lists of stdout and stderr for each execution,
                               and a single return code representing the
                               overall success of the apply.
    :rtype: tuple
    """
    log = common.configure_logging(__name__, log_level, log_file)
    if default_runtime:
        log.warning("DEPRECATION: 'default_runtime' does nothing, "
                    "use 'cont_cmd' instead")

    if cont_cmd == 'podman':
        r = runner.PodmanRunner(managed_by, cont_cmd=cont_cmd, log=log)
        builder = podman.PodmanBuilder(
            config_id=config_id,
            config=config,
            runner=r,
            labels=labels,
            log=log,
            cont_log_path=cont_log_path,
            healthcheck_disabled=healthcheck_disabled,
            cleanup=cleanup
        )
    else:
        r = runner.DockerRunner(managed_by, cont_cmd=cont_cmd, log=log)
        builder = compose1.ComposeV1Builder(
            config_id=config_id,
            config=config,
            runner=r,
            labels=labels,
            log=log,
            cleanup=cleanup
        )
    return builder.apply()


def cleanup(config_ids, managed_by, cont_cmd='podman', default_runtime=None,
            log_level=None, log_file=None):
    """Delete containers no longer applied, rename others to preferred name.

    :param list config_ids: List of config IDs still applied. All containers
                            managed by this tool will be deleted if their
                            config ID is not specified in this list.
    :param str managed_by: Name of the tool managing the containers. Only
                           containers labelled with this will be modified.
    :param str cont_cmd: Optional override to the container command to run.
    :param str default_runtime: (deprecated) does nothing.
    :param int log_level: optional log level for loggers
    :param int log_file: optional log file for messages
    """
    log = common.configure_logging(__name__, log_level, log_file)
    if default_runtime:
        log.warning("DEPRECATION: 'default_runtime' does nothing, "
                    "use 'cont_cmd' instead")

    if cont_cmd == 'podman':
        r = runner.PodmanRunner(managed_by, cont_cmd=cont_cmd, log=log)
    else:
        r = runner.DockerRunner(managed_by, cont_cmd=cont_cmd, log=log)

    r.delete_missing_configs(config_ids)
    r.rename_containers()


def list(managed_by, cont_cmd='podman', default_runtime=None,
         log_level=None, log_file=None):
    """List all containers associated with all config IDs.

    :param str managed_by: Name of the tool managing the containers. Only
                           containers labelled with this will be modified.
    :param str cont_cmd: Optional override to the container command to run.
    :param str default_runtime: (deprecated) does nothing.
    :param int log_level: optional log level for loggers
    :param int log_file: optional log file for messages

    :returns a dict where the key is the config ID and the value is a list of
             'podman inspect' dicts for each container.
    :rtype: defaultdict(list)
    """
    log = common.configure_logging(__name__, log_level, log_file)
    if default_runtime:
        log.warning("DEPRECATION: 'default_runtime' does nothing, "
                    "use 'cont_cmd' instead")

    if cont_cmd == 'podman':
        r = runner.PodmanRunner(managed_by, cont_cmd=cont_cmd, log=log)
    else:
        r = runner.DockerRunner(managed_by, cont_cmd=cont_cmd, log=log)

    return r.list_configs()


def debug(config_id, container_name, action, config, managed_by, labels=None,
          cont_cmd='podman', default_runtime=None, log_level=None,
          log_file=None):
    """Execute supplied container configuration.

    :param str config_id: Unique config ID, should not be re-used until any
                          running containers with that config ID have been
                          deleted.
    :param str container_name: Name of the container in the config you
                               wish to manipulate.
    :param str action: Action to take.
    :param dict config: Configuration data describing container actions to
                        apply.
    :param str managed_by: Name of the tool managing the containers. Only
                           containers labeled with this will be modified.
    :param dict labels: Optional keys/values of labels to apply to containers
                        created with this invocation.
    :param str cont_cmd: Optional override to the container command to run.
    :param str default_runtime: (deprecated) does nothing.
    :param int log_level: optional log level for loggers
    :param int log_file: optional log file for messages

    :returns integer return value from running command or failure for any
             other reason.
    :rtype: int
    """
    log = common.configure_logging(__name__, log_level, log_file)
    if default_runtime:
        log.warning("DEPRECATION: 'default_runtime' does nothing, "
                    "use 'cont_cmd' instead")

    if cont_cmd == 'podman':
        r = runner.PodmanRunner(managed_by, cont_cmd=cont_cmd, log=log)
        builder = podman.PodmanBuilder(
            config_id=config_id,
            config=config,
            runner=r,
            labels=labels,
            log=log
        )
    else:
        r = runner.DockerRunner(managed_by, cont_cmd=cont_cmd, log=log)
        builder = compose1.ComposeV1Builder(
            config_id=config_id,
            config=config,
            runner=r,
            labels=labels,
            log=log
        )
    if action == 'print-cmd':
        uname = r.unique_container_name(container_name)
        cmd = [
            r.cont_cmd,
            'run',
            '--name',
            uname
        ]
        builder.label_arguments(cmd, container_name)
        builder.container_run_args(cmd, container_name, uname)

        if '--health-cmd' in cmd:
            health_check_arg_index = cmd.index('--health-cmd') + 1

            # The argument given needs to be quoted to work properly with a
            # copy and paste of the full command.
            try:
                cmd[health_check_arg_index] = (
                    '"%s"' % cmd[health_check_arg_index])
            except IndexError:
                log.warning("No argument provided to --health-cmd.")

        print(' '.join(cmd))
    elif action == 'run':
        uname = r.unique_container_name(container_name)
        cmd = [
            r.cont_cmd,
            'run',
            '--name',
            uname
        ]
        builder.label_arguments(cmd, container_name)
        if builder.container_run_args(cmd, container_name, uname):
            return r.execute_interactive(cmd, log)
    elif action == 'dump-yaml':
        print(yaml.safe_dump(config, default_flow_style=False))
    elif action == 'dump-json':
        print(json.dumps(config, indent=4))
    else:
        raise ValueError('action should be one of: "dump-json", "dump-yaml"',
                         '"print-cmd", or "run"')


def delete(config_ids, managed_by, cont_cmd='podman', default_runtime=None,
           log_level=None, log_file=None):
    """Delete containers with the specified config IDs.

    :param list config_ids: List of config IDs to delete the containers for.
    :param str managed_by: Name of the tool managing the containers. Only
                           containers labelled with this will be modified.
    :param str cont_cmd: Optional override to the container command to run.
    :param str default_runtime: (deprecated) does nothing.
    """
    log = common.configure_logging(__name__, log_level, log_file)
    if default_runtime:
        log.warning("DEPRECATION: 'default_runtime' does nothing, "
                    "use 'cont_cmd' instead")

    if not config_ids:
        log.warn('No config IDs specified')

    if cont_cmd == 'podman':
        r = runner.PodmanRunner(managed_by, cont_cmd=cont_cmd, log=log)
    else:
        r = runner.DockerRunner(managed_by, cont_cmd=cont_cmd, log=log)

    for conf_id in config_ids:
        r.remove_containers(conf_id)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from paunch.builder import compose1
from paunch.tests import test_builder_base as tbb


class TestComposeV1Builder(tbb.TestBaseBuilder):
    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_cont_run_args(self, runner):
        config = {
            'one': {
                'image': 'centos:7',
                'privileged': True,
                'user': 'bar',
                'net': 'host',
                'ipc': 'host',
                'pid': 'container:bar',
                'uts': 'host',
                'restart': 'always',
                'healthcheck': {
                    'test': '/bin/true',
                    'interval': '30s',
                    'timeout': '10s',
                    'retries': 3
                },
                'env_file': '/tmp/foo.env',
                'log_tag': '{{.ImageName}}/{{.Name}}/{{.ID}}',
                'cpu_shares': 600,
                'cpuset_cpus': 'all',
                'mem_limit': '1G',
                'memswap_limit': '1G',
                'mem_swappiness': '60',
                'security_opt': 'label:disable',
                'cap_add': ['SYS_ADMIN', 'SETUID'],
                'cap_drop': ['NET_RAW'],
                'hostname': 'foohostname',
                'extra_hosts': [
                    'foohost:127.0.0.1',
                    'barhost:127.0.0.2'
                    ]

            }
        }
        runner.validate_volume_source.return_value = True
        builder = compose1.ComposeV1Builder('foo', config, runner)

        cmd = ['docker', 'run', '--name', 'one']
        builder.container_run_args(cmd, 'one')
        self.assertEqual(
            ['docker', 'run', '--name', 'one',
             '--detach=true', '--env-file=/tmp/foo.env',
             '--net=host', '--ipc=host', '--pid=container:bar',
             '--uts=host', '--health-cmd', '/bin/true',
             '--health-interval=30s',
             '--health-timeout=10s', '--health-retries=3',
             '--privileged=true', '--restart=always', '--user=bar',
             '--log-opt=tag={{.ImageName}}/{{.Name}}/{{.ID}}',
             '--cpu-shares=600',
             '--memory=1G',
             '--memory-swap=1G',
             '--memory-swappiness=60',
             '--security-opt=label:disable',
             '--hostname=foohostname',
             '--add-host=foohost:127.0.0.1',
             '--add-host=barhost:127.0.0.2',
             '--cap-add=SYS_ADMIN', '--cap-add=SETUID', '--cap-drop=NET_RAW',
             'centos:7'],
            cmd
        )

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_cont_run_args_validation_true(self, runner):
        config = {
            'one': {
                'image': 'foo',
                'volumes': ['/foo:/foo:rw', '/bar:/bar:ro'],
            }
        }
        runner.validate_volume_source.return_value = True
        builder = compose1.ComposeV1Builder('foo', config, runner)

        cmd = ['docker']
        self.assertTrue(builder.container_run_args(cmd, 'one'))
        self.assertEqual(
            ['docker', '--detach=true',
             '--volume=/foo:/foo:rw', '--volume=/bar:/bar:ro', 'foo'],
            cmd
        )

    @mock.patch('paunch.runner.DockerRunner', autospec=True)
    def test_cont_run_args_validation_false(self, runner):
        config = {
            'one': {
                'image': 'foo',
                'volumes': ['/foo:/foo:rw', '/bar:/bar:ro'],
            }
        }
        runner.validate_volume_source.return_value = False
        builder = compose1.ComposeV1Builder('foo', config, runner)

        cmd = ['docker']
        self.assertFalse(builder.container_run_args(cmd, 'one'))
        self.assertEqual(
            ['docker', '--detach=true',
             '--volume=/foo:/foo:rw', '--volume=/bar:/bar:ro', 'foo'],
            cmd
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import os
from paunch.builder import base
from paunch.utils import common


class PodmanBuilder(base.BaseBuilder):

    def __init__(self, config_id, config, runner, labels=None, log=None,
                 cont_log_path=None, healthcheck_disabled=False, cleanup=True):
        super(PodmanBuilder, self).__init__(config_id, config, runner,
                                            labels, log, cont_log_path,
                                            healthcheck_disabled, cleanup)

    def container_run_args(self, cmd, container, delegate=None):
        """Prepare the run command args, from the container configuration.

        :param cmd: The list of command options to be modified
        :param container: A dict with container configurations
        :param delegate: A predictable/unique name of the actual container
        :returns: True if configuration is valid, otherwise False
        """
        if delegate and container != delegate:
            self.log.debug("Container {} has a delegate "
                           "{}".format(container, delegate))
        if not delegate:
            delegate = container
        cconfig = self.config[container]

        # write out a pid file so we can restart the container delegate
        # via systemd
        cmd.append('--conmon-pidfile=/var/run/{}.pid'.format(delegate))

        # (bandini) - We must never use --detach when we call podman create as
        # it makes no sense whatsoever and podman-3.x errors out on it
        if 'create' not in cmd and cconfig.get('detach', True):
            cmd.append('--detach=true')

        if self.cont_log_path is not None:
            if os.path.isabs(self.cont_log_path):

                if not os.path.exists(self.cont_log_path):
                    os.makedirs(self.cont_log_path)
                log_path = os.path.join(self.cont_log_path, delegate)
                logging = ['--log-driver', 'k8s-file',
                           '--log-opt', 'path=%s.log' % log_path]
                cmd.extend(logging)
            else:
                raise ValueError('cont_log_path passed but not absolute.')

        self.list_or_string_arg(cconfig, cmd, 'env_file', '--env-file')
        self.list_or_dict_arg(cconfig, cmd, 'environment', '--env')
        self.boolean_arg(cconfig, cmd, 'remove', '--rm')
        self.boolean_arg(cconfig, cmd, 'interactive', '--interactive')
        self.boolean_arg(cconfig, cmd, 'tty', '--tty')
        self.string_arg(cconfig, cmd, 'net', '--net')
        self.string_arg(cconfig, cmd, 'ipc', '--ipc')
        self.string_arg(cconfig, cmd, 'pid', '--pid')
        self.string_arg(cconfig, cmd, 'uts', '--uts')
        self.string_arg(cconfig, cmd, 'pids_limit', '--pids-limit')
        # TODO(sbaker): implement ulimits property, deprecate this ulimit
        # property
        for u in cconfig.get('ulimit', []):
            if u:
                cmd.append('--ulimit=%s' % u)

        self.string_arg(cconfig, cmd, 'privileged', '--privileged', self.lower)
        self.string_arg(cconfig, cmd, 'user', '--user')
        self.list_arg(cconfig, cmd, 'group_add', '--group-add')
        self.list_arg(cconfig, cmd, 'volumes', '--volume')
        self.list_arg(cconfig, cmd, 'volumes_from', '--volumes-from')
        # TODO(sbaker): deprecate log_tag, implement log_driver, log_opt
        if 'log_tag' in cconfig:
            cmd.append('--log-opt=tag=%s' % cconfig['log_tag'])
        self.string_arg(cconfig, cmd, 'cpu_shares', '--cpu-shares')
        self.string_arg(cconfig, cmd, 'mem_limit', '--memory')
        self.string_arg(cconfig, cmd, 'memswap_limit', '--memory-swap')
        self.string_arg(cconfig, cmd, 'mem_swappiness', '--memory-swappiness')
        self.list_or_string_arg(cconfig, cmd, 'security_opt', '--security-opt')
        self.string_arg(cconfig, cmd, 'stop_signal', '--stop-signal')
        self.string_arg(cconfig, cmd, 'hostname', '--hostname')
        for extra_host in cconfig.get('extra_hosts', []):
            if extra_host:
                cmd.append('--add-host=%s' % extra_host)
        if 'cpuset_cpus' in cconfig:
            # 'all' is a special value to directly configure all CPUs
            # that are available. Without specifying --cpuset-cpus, we'll
            # let the container engine to figure out what CPUs are online.
            # https://bugs.launchpad.net/tripleo/+bug/1868135
            # https://bugzilla.redhat.com/show_bug.cgi?id=1813091
            if cconfig['cpuset_cpus'] != 'all':
                cmd.append('--cpuset-cpus=%s' % cconfig['cpuset_cpus'])
        else:
            with open('/proc/cmdline') as cmdline:
                if 'isolcpus' in cmdline.read():
                    cmd.append('--cpuset-cpus=%s' %
                               common.get_cpus_allowed_list())

        self.string_arg(cconfig, cmd,
                        'stop_grace_period', '--stop-timeout',
                        self.duration)

        self.list_arg(cconfig, cmd, 'cap_add', '--cap-add')
        self.list_arg(cconfig, cmd, 'cap_drop', '--cap-drop')
        self.string_arg(cconfig, cmd, 'check_interval', '--check-interval')

        cmd.append(cconfig.get('image', ''))
        cmd.extend(self.command_argument(cconfig.get('command')))

        return self.validate_volumes(cconfig.get('volumes', []))
